#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <shellapi.h>
#include <shlobj.h>
#include <richedit.h>
#include <commctrl.h>
#include <wchar.h>
#include "akelpad.h"
#include "edit.h"

#define xatoi
#define hex2dec
#include "ConvFunc.h"


//Process
extern HANDLE hHeap;

//WinMain
extern HINSTANCE hInstance;
extern char *lpCmdLine;
extern int nCmdShow;
extern STARTUPINFO lpStartupInfo;

//Versions
extern BOOL bOldWindows;
extern BOOL bOldRichEdit;
extern BOOL bOldComctl32;

//Buffers
extern char buf[BUFFER_SIZE];
extern char buf2[BUFFER_SIZE];
extern unsigned char pcTranslateBuffer[TRANSLATE_BUFFER_SIZE];

//Language
extern char szLangModule[MAX_PATH];
extern HMODULE hLangLib;
extern int nLangModules;
extern LANGID wSystemLanguage;

//Main Window
extern HWND hMainWnd;
extern HWND hWndEdit;
extern HWND hSyncWindow;
extern HWND hStatus;
extern HWND hDlgModeless;
extern RECT rcMainWindow;
extern int nMainStyle;
extern int nLastMainSize;
extern int nStatusHeight;
extern HACCEL hMainAccel;
extern HICON hMainIcon;
extern HICON hIconEmpty;
extern HMENU hMainMenu;
extern HMENU hPopupMenu;
extern HMENU hPopupEdit;
extern HMENU hPopupView;

//Codepages
extern int codepages[65536];
extern BOOL bEnumCodePages;
extern BOOL bDefaultBOM;
extern BOOL bCurrentBOM;
extern int nRecodeCodePage;
extern int nCurrentCodePage;
extern int nDefaultCodePage;
extern int nAnsiCodePage;
extern int nOemCodePage;
extern BOOL (WINAPI *GetCPInfoExPtr) (UINT, DWORD, LPCPINFOEX);

//Recent files
extern char szCurrentFile[MAX_PATH];
extern int nRecentFiles;
extern char (*szRecentNames)[MAX_PATH];
extern DWORD *szRecentPositions;
extern DWORD *szRecentCodepages;

//Open/Save dialog
extern OPENFILENAME ofn;
extern char szFileBuffer[FILELIST_BUFFER_SIZE];
extern char szFilter[MAX_PATH];
extern BOOL bAutodetect;
extern BOOL bSaveDlg;
extern DWORD dwOfnFlags;
extern BOOL bOfnBOM;
extern int nOfnCodePage;
extern BOOL bNewLineDetected;
extern WNDPROC OldPreviewProc;

//Find/Replace dialog
extern DWORD ftflags;
extern BOOL bRecodeDlg;
extern BOOL bReplaceDlg;
extern int nReplaceCount;
extern int nSearchStrings;
extern wchar_t *wszFind_orig;
extern wchar_t *wszReplace_orig;
extern wchar_t *wszFind;
extern wchar_t *wszReplace;

//Options dialog
extern PROPSHEETHEADER psh;
extern PROPSHEETPAGE psp[2];
extern HHOOK cbtHook;

//Font/Color
extern LOGFONT logfont;
extern CHOOSEFONT cf;
extern CHOOSECOLOR cc;
extern COLORREF crBackground;
extern COLORREF crFont;
extern COLORREF crCustColors[16];

//Print
extern PAGESETUPDLG psdPage;
extern PRINTDLG pd;
extern BOOL bGlobalPrint;

//Edit position
extern CHARRANGE chrg;
extern int nPosition;
extern int nPositionOld;
extern int nMaxChars;
extern int nMcstep;

//Edit state
extern BOOL bModified;
extern BOOL bInsertState;
extern BOOL bUnixNewLine;
extern BOOL bWordWrap;
extern BOOL bOnTop;
extern BOOL bStatusBar;
extern BOOL bSaveTime;
extern BOOL bSingleOpenFile;
extern BOOL bSingleOpenProgram;
extern BOOL bKeepSpace;
extern DWORD dwEditMargins;

//Word breaking
extern wchar_t wszDelimiters[DELIMITERS_SIZE];
extern BOOL bFirstWordBreak;

//Execute
extern char szCommand[BUFFER_SIZE];
extern char szExeDir[MAX_PATH];

//Associations
extern char szFileTypesOpen[MAX_PATH];
extern char szFileTypesEdit[MAX_PATH];
extern char szFileTypesPrint[MAX_PATH];
extern DWORD dwFileTypesAssociated;

//Registered messages
extern UINT uiOpenOnStartMessage;

//Mdi
extern BOOL bMDI;
extern BOOL bRegMDI;
extern HWND hMdiClient;
extern HWND hWndMdiActive;
extern HWND hWndMdiDestroyed;
extern BOOL bMdiMaximize;
extern BOOL bMdiNoWindows;
extern BOOL bMdiReopen;
extern HWND hTab;
extern HFONT hTabFont;
extern LOGFONT tabfont;
extern int nTabView;
extern int nTabType;
extern HIMAGELIST hImageList;
extern BOOL bFileExitError;
extern int nMdiStyle;
extern WNDPROC OldMdiClientProc;
extern WNDPROC OldTabProc;
extern MDIWINDOWDATA *lpWindowData;
extern UINT (WINAPI *ExtractIconExWPtr) (LPCWSTR, int, HICON *, HICON *, UINT);


//// Init function

void CreateEditWindow(HWND hWnd)
{
  RECT rcRect;
  int nLangoptions;
  BOOL bModifiedTmp=bModified;

  GetClientRect(hWnd, &rcRect);

  hWndEdit=CreateWindow("richedit20a",
                        NULL,
                        WS_CHILD|WS_VISIBLE|WS_HSCROLL|WS_VSCROLL|
                          ES_LEFT|ES_MULTILINE|ES_AUTOHSCROLL|ES_AUTOVSCROLL|
                          ES_DISABLENOSCROLL|ES_SUNKEN|ES_NOHIDESEL,
                        0, 0, rcRect.right, rcRect.bottom - (bStatusBar?nStatusHeight:0),
                        hWnd,
                        (HMENU)ID_EDIT,
                        hInstance,
                        NULL);

  SendMessage(hWndEdit, EM_EXLIMITTEXT, 0, nMaxChars);
  nLangoptions=SendMessage(hWndEdit, EM_GETLANGOPTIONS, 0, 0);
  nLangoptions&=~IMF_AUTOFONT;
  nLangoptions&=~IMF_AUTOFONTSIZEADJUST;
  SendMessage(hWndEdit, EM_SETLANGOPTIONS, 0, nLangoptions);
  SendMessage(hWndEdit, EM_SETTEXTMODE, TM_RICHTEXT|TM_MULTILEVELUNDO|TM_MULTICODEPAGE, 0);
  SendMessage(hWndEdit, EM_SETWORDBREAKPROC, 0, (LPARAM)EditWordBreakProc);
  SendMessage(hWndEdit, EM_SETEVENTMASK, 0, ENM_SELCHANGE|ENM_CHANGE|ENM_KEYEVENTS|ENM_MOUSEEVENTS);

  SetWindowText(hWndEdit, "");
  DoViewWordWrap(hWndEdit, bWordWrap, TRUE);
  SendMessage(hWndEdit, EM_SETMARGINS, EC_LEFTMARGIN|EC_RIGHTMARGIN, dwEditMargins);
  SendMessage(hWndEdit, EM_SETBKGNDCOLOR, 0, crBackground);
  SetChosenFont(hWndEdit, logfont);
  SetChosenFontColor(hWndEdit, crFont);
  SetModify(bModifiedTmp, FALSE);
}


//// Menu functions

BOOL DoFileNew()
{
  if (!bMDI)
  {
    if (!SaveChanged()) return FALSE;

    if (nRecentFiles && *szCurrentFile)
    {
      RecentFilesZero();
      RecentFilesRead();
      RecentFilesUpdate(szCurrentFile, nPosition, MAKELONG(nCurrentCodePage, bCurrentBOM));
      RecentFilesSave();
    }

    HideCaret(NULL);
    SetWindowText(hWndEdit, "");
    ShowCaret(NULL);
    szCurrentFile[0]='\0';
    SetWindowText(hMainWnd, APP_SHORT_TITLE);
    SetUnixNewLine(FALSE, FALSE);
    SetModify(FALSE, FALSE);
    SetCodePage(MAKELONG(nDefaultCodePage, bDefaultBOM), FALSE);
  }
  else
  {
    if (hWndMdiActive) SendMessage(hMdiClient, WM_MDIGETACTIVE, 0, (LPARAM)&bMdiMaximize);
    CreateMDIWindow(APP_MDI_NAME, "", (bMdiMaximize == TRUE)?WS_MAXIMIZE:0, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, hMdiClient, hInstance, 0);
  }
  return TRUE;
}

void DoFileNewWindow()
{
  STARTUPINFO si={0};
  PROCESS_INFORMATION pi={0};

  if (!GetModuleFileName(hInstance, buf, MAX_PATH)) return;

  si.cb=sizeof(STARTUPINFO);
  si.dwFlags=STARTF_USESHOWWINDOW;
  si.wShowWindow=SW_SHOWNORMAL;

  if (CreateProcess(buf, NULL, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
  {
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
  }
}

BOOL DoFileOpen()
{
  if (!bMDI && !SaveChanged()) return FALSE;

  bSaveDlg=FALSE;
  ofn.lpstrDefExt=NULL;
  if (!bMDI)
    ofn.Flags&=~OFN_ALLOWMULTISELECT;
  else
    ofn.Flags|=OFN_ALLOWMULTISELECT;
  lstrcpy(szFileBuffer, szCurrentFile);

  if (GetOpenFileName(&ofn))
  {
    if (!bMDI)
    {
      if (OpenDocument(szFileBuffer, dwOfnFlags, nOfnCodePage, bOfnBOM) < 0)
        return FALSE;
    }
    else
    {
      char szFile[MAX_PATH];
      char *pFile=szFileBuffer + lstrlen(szFileBuffer) + 1;

      if (*pFile)
      {
        if (*(pFile - 2) == '\\') *(pFile - 2)='\0';

        do
        {
          if (IsPathFull(pFile))
            lstrcpy(szFile, pFile);  //.lnk target
          else
            wsprintf(szFile, "%s\\%s", szFileBuffer, pFile);
          OpenDocument(szFile, dwOfnFlags, nOfnCodePage, bOfnBOM);
        }
        while (*(pFile+=lstrlen(pFile) + 1));

        return TRUE;
      }
      else
      {
        if (OpenDocument(szFileBuffer, dwOfnFlags, nOfnCodePage, bOfnBOM) < 0)
          return FALSE;
      }
    }
  }
  return FALSE;
}

void DoFileReopenAs(DWORD Flags, int nCodePage, BOOL bBOM)
{
  int i;

  if (!*szCurrentFile) return;

  if (bModified)
  {
    API_LoadString(hLangLib, MSG_FILE_WILL_BE_REOPENED, buf, BUFFER_SIZE);
    i=MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OKCANCEL|MB_ICONEXCLAMATION);
  }
  if (!bModified || i == IDOK)
  {
    bMdiReopen=TRUE;
    OpenDocument(szCurrentFile, Flags, nCodePage, bBOM);
    bMdiReopen=FALSE;
  }
}

BOOL DoFileSave()
{
  if (!bModified && *szCurrentFile) return FALSE;
  if (!*szCurrentFile) return DoFileSaveAs();
  SaveDocument(szCurrentFile, nCurrentCodePage, bCurrentBOM);
  return TRUE;
}

BOOL DoFileSaveAs()
{
  bSaveDlg=TRUE;
  ofn.lpstrDefExt="txt";
  ofn.Flags&=~OFN_ALLOWMULTISELECT;
  lstrcpy(szFileBuffer, szCurrentFile);

  if (GetSaveFileName(&ofn))
  {
    if (SaveDocument(szFileBuffer, nOfnCodePage, bOfnBOM))
      return TRUE;
  }
  return FALSE;
}

void DoFilePageSetup()
{
  PageSetupDlg(&psdPage);
}

void DoFilePrint()
{
  FORMATRANGE fr;
  DOCINFO di={0};
  int nextch;
  int width;
  int height;

  pd.hDevMode=psdPage.hDevMode;
  pd.hDevNames=psdPage.hDevNames;

  API_LoadString(hLangLib, STR_DOCNAME, buf, BUFFER_SIZE);
  di.cbSize=sizeof(DOCINFO);
  di.lpszDocName=(*szCurrentFile)?szCurrentFile:buf;

  if (!PrintDlg(&pd)) return;

  width=GetDeviceCaps(pd.hDC, HORZSIZE)*56;  //about 56 twips per millimeter
  height=GetDeviceCaps(pd.hDC, VERTSIZE)*56;

  fr.hdc=pd.hDC;
  fr.hdcTarget=pd.hDC;

  if (psdPage.Flags & PSD_INHUNDREDTHSOFMILLIMETERS)
  {
    fr.rc.left=psdPage.rtMargin.left*567/1000;
    fr.rc.top=psdPage.rtMargin.top*567/1000;
    fr.rc.right=width-psdPage.rtMargin.right*567/1000;
    fr.rc.bottom=height-psdPage.rtMargin.bottom*567/1000;
  }
  else
  {
    fr.rc.left=psdPage.rtMargin.left*144/100;
    fr.rc.top=psdPage.rtMargin.top*144/100;
    fr.rc.right=width-psdPage.rtMargin.right*144/100;
    fr.rc.bottom=height-psdPage.rtMargin.bottom*144/100;
  }
  fr.rcPage.left=0;
  fr.rcPage.top=0;
  fr.rcPage.right=width;
  fr.rcPage.bottom=height;

  if (pd.Flags & PD_PRINTTOFILE) di.lpszOutput="FILE:";

  if (StartDoc(pd.hDC, &di) <= 0)
  {
    API_LoadString(hLangLib, MSG_ERROR_PRINT, buf, BUFFER_SIZE);
    MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
    return;
  }

  if (pd.Flags & PD_SELECTION)
  {
    SendMessage(hWndEdit, EM_EXGETSEL, 0, (LPARAM)&fr.chrg);
  }
  else
  {
    fr.chrg.cpMin=0;
    fr.chrg.cpMax=-1;
  }

  do
  {
    nextch=SendMessage(hWndEdit, EM_FORMATRANGE, FALSE, (LPARAM)&fr);
    if (nextch == fr.chrg.cpMin) break;
    fr.chrg.cpMin=nextch;
    StartPage(pd.hDC);
    if (!SendMessage(hWndEdit, EM_DISPLAYBAND, 0, (LPARAM)&fr.rc))
    {
      EndPage(pd.hDC);
      API_LoadString(hLangLib, MSG_ERROR_PRINT, buf, BUFFER_SIZE);
      MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
      break;
    }
    EndPage(pd.hDC);
  }
  while (true);

  EndDoc(pd.hDC);
  DeleteDC(pd.hDC);
}

BOOL DoFileExit()
{
  if (!SaveChanged()) return FALSE;

  if (nRecentFiles && *szCurrentFile)
  {
    RecentFilesZero();
    RecentFilesRead();
    RecentFilesUpdate(szCurrentFile, nPosition, MAKELONG(nCurrentCodePage, bCurrentBOM));
    RecentFilesSave();
  }
  return TRUE;
}

void DoEditUndo(HWND hWnd)
{
  SendMessage(hWnd, EM_UNDO, 0, 0);
}

void DoEditRedo(HWND hWnd)
{
  SendMessage(hWnd, EM_REDO, 0, 0);
}

void DoEditCut(HWND hWnd)
{
  SendMessage(hWnd, WM_CUT, 0, 0);
}

void DoEditCopy(HWND hWnd)
{
  SendMessage(hWnd, WM_COPY, 0, 0);
}

void DoEditPaste(HWND hWnd)
{
  HGLOBAL hData;
  LPVOID pData;

  if (!OpenClipboard(hWnd))
    return;

  if ((hData=GetClipboardData(CF_UNICODETEXT)) && (pData=GlobalLock(hData)))
  {
    ReplaceSelW(hWnd, (wchar_t *)pData);
    GlobalUnlock(hData);
  }
  else if ((hData=GetClipboardData(CF_TEXT)) && (pData=GlobalLock(hData)))
  {
    ReplaceSelA(hWnd, (char *)pData);
    GlobalUnlock(hData);
  }
  CloseClipboard();
}

void DoEditPasteANSI(HWND hWnd)
{
  HGLOBAL hData;
  LPVOID pData;

  if (!OpenClipboard(hWnd))
    return;

  if ((hData=GetClipboardData(CF_TEXT)) && (pData=GlobalLock(hData)))
  {
    ReplaceSelA(hWnd, (char *)pData);
    GlobalUnlock(hData);
  }
  CloseClipboard();
}

void DoEditClear(HWND hWnd)
{
  SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)"");
}

void DoEditSelectAll(HWND hWnd)
{
  SendMessage(hWnd, EM_SETSEL, 0, -1);
}

void DoEditInsertChar()
{
  ShellExecute(hMainWnd, "open", "charmap.exe", NULL, NULL, SW_SHOWDEFAULT);
}

void DoEditRecode(HWND hWnd)
{
  bRecodeDlg=TRUE;
  API_DialogBoxA(hLangLib, MAKEINTRESOURCE(IDD_RECODE), hMainWnd, (DLGPROC)RecodeDlgProc);
  if (bRecodeDlg)
  {
    RecodeText(hWnd, nRecodeCodePage);
    bRecodeDlg=FALSE;
  }
}

BOOL DoEditInsertCharInSelection(HWND hWnd, int nAction, char ch)
{
  char *pText;
  TEXTRANGE txtrng;
  int nLayout;
  int a, b, c;
  int nTextSize;
  int nTabsNumber;
  int nMinLine;
  int nMaxLine;
  int nMinLineIndex;
  int nMaxLineIndex;
  int nMaxLineLength;
  BOOL bCrLessEnd=FALSE;

  if (chrg.cpMin != chrg.cpMax)
  {
    txtrng.chrg.cpMin=chrg.cpMin;
    txtrng.chrg.cpMax=chrg.cpMax;
    nMinLine=SendMessage(hWnd, EM_LINEFROMCHAR, txtrng.chrg.cpMin, 0);
    nMaxLine=SendMessage(hWnd, EM_LINEFROMCHAR, txtrng.chrg.cpMax, 0);
    nMinLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMinLine, 0);
    nMaxLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMaxLine, 0);
    nMaxLineLength=SendMessage(hWnd, EM_LINELENGTH, txtrng.chrg.cpMax, 0);

    if (nMinLine != nMaxLine || (nMinLineIndex == txtrng.chrg.cpMin && !nMaxLineLength))
    {
      if (nAction == CHECK_STATE) return TRUE;
      if (nMaxLineIndex != txtrng.chrg.cpMax)
      {
        if (nMaxLineLength) txtrng.chrg.cpMax=nMaxLineIndex + nMaxLineLength + 1;
        else bCrLessEnd=TRUE;
      }
      else --nMaxLine;

      txtrng.chrg.cpMin=nMinLineIndex;
      SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&txtrng.chrg);

      if (nAction == CHAR_INSERT)
      {
        nTabsNumber=nMaxLine - nMinLine + 1;
        nTextSize=txtrng.chrg.cpMax - txtrng.chrg.cpMin + 1;

        if (pText=(char *)API_HeapAlloc(hHeap, 0, nTextSize + nTabsNumber + 1))
        {
          nLayout=SetLayout();
          txtrng.lpstrText=pText + nTabsNumber;
          SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
          RestoreLayout(nLayout);

          for (a=1, b=nTabsNumber, c=1, pText[0]=ch; c < nTabsNumber; pText[a++]=pText[b++])
            while (pText[b] == '\r' && c < nTabsNumber)
            {
              pText[a++]=pText[b++];
              pText[a++]=ch;
              ++c;
            }
        }
        else return FALSE;
      }
      else if (nAction == CHAR_DELETE)
      {
        nTextSize=txtrng.chrg.cpMax - txtrng.chrg.cpMin + 1;

        if (pText=(char *)API_HeapAlloc(hHeap, 0, nTextSize + 2))
        {
          nLayout=SetLayout();
          txtrng.lpstrText=pText + 1;
          SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
          RestoreLayout(nLayout);

          for (a=0, b=(pText[1] == ch)?2:1; b <= nTextSize; pText[a++]=pText[b++])
            while (pText[b] == '\r')
            {
              pText[a++]=pText[b++];
              if (pText[b] == ch) ++b;
            }
        }
        else return FALSE;
      }
      SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)pText);
      if (bCrLessEnd) txtrng.chrg.cpMax=chrg.cpMax + 1;
      else txtrng.chrg.cpMax=chrg.cpMax;
      SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&txtrng.chrg);
      API_HeapFree(hHeap, 0, (LPVOID)pText);
      return TRUE;
    }
  }
  if (nAction == CHECK_STATE) return FALSE;
  buf[0]=ch;
  buf[1]='\0';
  SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)buf);
  return TRUE;
}

BOOL DoEditDeleteTrailingWhitespaces(HWND hWnd, int nAction)
{
  char *pText;
  TEXTRANGE txtrng;
  int nLayout;
  int a, b;
  int nTextSize;
  int nMaxLine;
  int nMaxLineIndex;
  int nMaxLineLength;
  BOOL bCrLessEnd=FALSE;

  if (chrg.cpMin != chrg.cpMax)
  {
    if (nAction == CHECK_TRAILING_WHITESPACES) return TRUE;
    txtrng.chrg.cpMin=chrg.cpMin;
    txtrng.chrg.cpMax=chrg.cpMax;
    nMaxLine=SendMessage(hWnd, EM_LINEFROMCHAR, txtrng.chrg.cpMax, 0);
    nMaxLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMaxLine, 0);
    nMaxLineLength=SendMessage(hWnd, EM_LINELENGTH, txtrng.chrg.cpMax, 0);

    if (nAction == DELETE_TRAILING_WHITESPACES)
    {
      if (nMaxLineIndex != txtrng.chrg.cpMax && !nMaxLineLength)
        bCrLessEnd=TRUE;
      nTextSize=txtrng.chrg.cpMax - txtrng.chrg.cpMin + 1;

      if (pText=(char *)API_HeapAlloc(hHeap, 0, nTextSize + 2))
      {
        nLayout=SetLayout();
        txtrng.lpstrText=pText + 1;
        SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
        RestoreLayout(nLayout);

        for (a=0, b=1; b <= nTextSize; pText[a++]=pText[b++])
          while (pText[b] == '\r' || pText[b] == '\0')
          {
            do --a;
            while (a >= 0 && (pText[a] == '\t' || pText[a] == ' '));

            ++a;
            pText[a++]=pText[b++];
          }
      }
      else return FALSE;

      SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)pText);
      if (bCrLessEnd) txtrng.chrg.cpMax=chrg.cpMax + 1;
      else txtrng.chrg.cpMax=chrg.cpMax;
      SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&txtrng.chrg);
      API_HeapFree(hHeap, 0, (LPVOID)pText);
      return TRUE;
    }
  }
  if (nAction == CHECK_TRAILING_WHITESPACES) return FALSE;
  SendMessage(hWnd, WM_KEYDOWN, VK_BACK, 0);
  return TRUE;
}

void DoEditChangeCase(HWND hWnd, int nCase)
{
  TEXTRANGE txtrng;
  int nTextSize;
  char *pStart;
  char *pEnd;
  int nLayout;

  if (chrg.cpMin != chrg.cpMax)
  {
    txtrng.chrg.cpMin=chrg.cpMin;
    txtrng.chrg.cpMax=chrg.cpMax;
    nTextSize=txtrng.chrg.cpMax - txtrng.chrg.cpMin;

    pStart=txtrng.lpstrText=(char *)API_HeapAlloc(hHeap, 0, nTextSize + 1);
    if (!txtrng.lpstrText) return;
    nLayout=SetLayout();

    pEnd=pStart + nTextSize;
    SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);

    if (nCase == UPPERCASE)
    {
      CharUpperBuff(txtrng.lpstrText, nTextSize);
    }
    else if (nCase == LOWERCASE)
    {
      CharLowerBuff(txtrng.lpstrText, nTextSize);
    }
    else if (nCase == SENTENCECASE)
    {
      while (pStart < pEnd)
      {
        while (pStart < pEnd && (*pStart == ' ' || *pStart == '\t' || *pStart == '\r' || strchr(SENTENCE_SELIMITERS, *pStart)))
          ++pStart;

        if (pStart < pEnd) *pStart++=(char)CharUpper((char*)(unsigned)(unsigned char)*pStart);

        while (pStart < pEnd && !strchr(SENTENCE_SELIMITERS, *pStart))
          *pStart++=(char)CharLower((char*)(unsigned)(unsigned char)*pStart);
      }
    }
    else if (nCase == TITLECASE)
    {
      while (pStart < pEnd)
      {
        while (pStart < pEnd && strchr(TITLE_DELIMITERS, *pStart))
          ++pStart;

        if (pStart < pEnd) *pStart++=(char)CharUpper((char*)(unsigned)(unsigned char)*pStart);

        while (pStart < pEnd && !strchr(TITLE_DELIMITERS, *pStart))
          *pStart++=(char)CharLower((char*)(unsigned)(unsigned char)*pStart);
      }
    }
    else if (nCase == INVERTCASE)
    {
      while (pStart < pEnd)
      {
        if (IsCharLower(*pStart))
          *pStart++=(char)CharUpper((char*)(unsigned)(unsigned char)*pStart);
        else
          *pStart++=(char)CharLower((char*)(unsigned)(unsigned char)*pStart);
      }
    }
    SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)txtrng.lpstrText);
    SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&txtrng.chrg);
    API_HeapFree(hHeap, 0, (LPVOID)txtrng.lpstrText);
    RestoreLayout(nLayout);
  }
}

void DoEditFind()
{
  if (!hDlgModeless)
  {
    bReplaceDlg=FALSE;
    if (ftflags & FR_SELECTION) ftflags&=~FR_SELECTION;

    if (bOldWindows && (hDlgModeless=API_CreateDialogA(hLangLib, MAKEINTRESOURCE(IDD_FIND), hMainWnd, (DLGPROC)FindAndReplaceDlgProcW)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
    else if (!bOldWindows && (hDlgModeless=API_CreateDialogW(hLangLib, MAKEINTRESOURCEW(IDD_FIND), hMainWnd, (DLGPROC)FindAndReplaceDlgProcW)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
  }
}

void DoEditFindNextDown(HWND hWnd)
{
  DWORD tmpflags=ftflags;

  if (wszFind)
  {
    ftflags&=~FR_UP&~FR_BEGINNING&~FR_SELECTION;
    ftflags|=FR_DOWN;

    FindTextW(hWnd);

    ftflags=tmpflags;
  }
  else DoEditFind();
}

void DoEditFindNextUp(HWND hWnd)
{
  DWORD tmpflags=ftflags;

  if (wszFind)
  {
    ftflags&=~FR_DOWN&~FR_BEGINNING&~FR_SELECTION;
    ftflags|=FR_UP;

    FindTextW(hWnd);

    ftflags=tmpflags;
  }
  else DoEditFind();
}

void DoEditReplace()
{
  if (!hDlgModeless)
  {
    bReplaceDlg=TRUE;

    if (bOldWindows && (hDlgModeless=API_CreateDialogA(hLangLib, MAKEINTRESOURCE(IDD_REPLACE), hMainWnd, (DLGPROC)FindAndReplaceDlgProcW)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
    else if (!bOldWindows && (hDlgModeless=API_CreateDialogW(hLangLib, MAKEINTRESOURCEW(IDD_REPLACE), hMainWnd, (DLGPROC)FindAndReplaceDlgProcW)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
  }
}

void DoEditGoToLine()
{
  if (!hDlgModeless)
  {
    if (bOldWindows && (hDlgModeless=API_CreateDialogA(hLangLib, MAKEINTRESOURCE(IDD_GOTOLINE), hMainWnd, (DLGPROC)GoToLineDlgProc)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
    else if (!bOldWindows && (hDlgModeless=API_CreateDialogW(hLangLib, MAKEINTRESOURCEW(IDD_GOTOLINE), hMainWnd, (DLGPROC)GoToLineDlgProc)))
    {
      ShowWindow(hDlgModeless, SW_SHOW);
    }
  }
}

void DoViewFont(HWND hWnd)
{
  LOGFONT tmp;

  memcpy(&tmp, &logfont, sizeof(LOGFONT));

  if (ChooseFont(&cf))
  {
    SetChosenFont(hWnd, logfont);
  }
  else memcpy(&logfont, &tmp, sizeof(LOGFONT));
}

void DoViewTextColor(HWND hWnd)
{
  COLORREF tmp=cc.rgbResult=crFont;

  if (ChooseColor(&cc))
  {
    crFont=cc.rgbResult;
    SetChosenFontColor(hWnd, crFont);
  }
  else crFont=tmp;
}


void DoViewBGColor(HWND hWnd)
{
  COLORREF tmp=cc.rgbResult=crBackground;

  if (ChooseColor(&cc))
  {
    crBackground=cc.rgbResult;
    SendMessage(hWnd, EM_SETBKGNDCOLOR, 0, crBackground);
  }
  else crBackground=tmp;
}

void DoViewFontSize(HWND hWnd, int nAction)
{
  if (nAction == DECREASE_FONT)
  {
    if (logfont.lfHeight >= 0)
      logfont.lfHeight-=1;
    else
      logfont.lfHeight+=1;
    SetChosenFontSize(hWnd, logfont.lfHeight);
  }
  else if (nAction == INCREASE_FONT)
  {
    if (logfont.lfHeight >= 0)
      logfont.lfHeight+=1;
    else
      logfont.lfHeight-=1;
    SetChosenFontSize(hWnd, logfont.lfHeight);
  }
}

void DoViewAlignment(HWND hWnd, WORD wAlignment)
{
  PARAFORMAT2 pf;

  pf.cbSize=sizeof(PARAFORMAT2);
  pf.dwMask=PFM_ALIGNMENT;
  pf.wAlignment=wAlignment;
  SendMessage(hWnd, EM_SETPARAFORMAT, 0, (LPARAM)&pf);
}

void DoViewWordWrap(HWND hWnd, BOOL bState, BOOL bFirst)
{
  NMHDR nmhdr={hWnd, ID_EDIT, EN_SELCHANGE};

  CheckMenuItem(hMainMenu, IDM_VIEW_WORDWRAP, bState?MF_CHECKED:MF_UNCHECKED);
  if (bFirst != TRUE && bState == bWordWrap) return;
  if (bFirst == TRUE && bState == FALSE) return;

  if (bWordWrap=bState)
    SendMessage(hWnd, EM_SETTARGETDEVICE, NULL, 0);
  else
    SendMessage(hWnd, EM_SETTARGETDEVICE, NULL, -1);
  SendMessage(hMainWnd, WM_NOTIFY, ID_EDIT, (LPARAM)&nmhdr);
}

void DoViewOnTop(BOOL bState, BOOL bFirst)
{
  CheckMenuItem(hMainMenu, IDM_VIEW_ONTOP, bState?MF_CHECKED:MF_UNCHECKED);
  if (bFirst != TRUE && bState == bOnTop) return;

  if (bOnTop=bState)
    SetWindowPos(hMainWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
  else
    SetWindowPos(hMainWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
}

void DoViewShowStatusBar(BOOL bState, BOOL bFirst)
{
  RECT rcRect;

  CheckMenuItem(hMainMenu, IDM_VIEW_SHOW_STATUSBAR, bState?MF_CHECKED:MF_UNCHECKED);
  if (bFirst != TRUE && bState == bStatusBar) return;
  bStatusBar=bState;

  ShowWindow(hStatus, bStatusBar?SW_SHOW:SW_HIDE);
  GetClientRect(hMainWnd, &rcRect);
  UpdateSize(rcRect.right, rcRect.bottom);
}

void DoSettingsExec(char *pCommand)
{
  STARTUPINFO si={0};
  PROCESS_INFORMATION pi={0};
  char *szResult;
  int nLen;
  int i;

  nLen=ParseExecCommand(pCommand, NULL, 0) + 1;

  if (szResult=(char *)API_HeapAlloc(hHeap, 0, nLen + 1))
  {
    ParseExecCommand(pCommand, szResult, nLen);
    si.cb=sizeof(STARTUPINFO);
    if (i=CreateProcess(NULL, szResult, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi))
    {
      CloseHandle(pi.hProcess);
      CloseHandle(pi.hThread);
    }
    else
    {
      API_LoadString(hLangLib, MSG_ERROR_RUN, buf, BUFFER_SIZE);
      MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONEXCLAMATION);
    }
    API_HeapFree(hHeap, 0, (LPVOID)szResult);
  }
}

void DoSettingsSaveTime(BOOL bState)
{
  CheckMenuItem(hMainMenu, IDM_OPTIONS_SAVETIME, bState?MF_CHECKED:MF_UNCHECKED);
  bSaveTime=bState;
}

void DoSettingsKeepSpace(BOOL bState)
{
  CheckMenuItem(hMainMenu, IDM_OPTIONS_KEEPSPACE, bState?MF_CHECKED:MF_UNCHECKED);
  bKeepSpace=bState;
}

void DoSettingsSingleOpenFile(BOOL bState)
{
  CheckMenuItem(hMainMenu, IDM_OPTIONS_SINGLEOPEN_FILE, bState?MF_CHECKED:MF_UNCHECKED);
  bSingleOpenFile=bState;
}

void DoSettingsSingleOpenProgram(BOOL bState)
{
  CheckMenuItem(hMainMenu, IDM_OPTIONS_SINGLEOPEN_PROGRAM, bState?MF_CHECKED:MF_UNCHECKED);
  bSingleOpenProgram=bState;
}

void DoSettingsOptions()
{
  API_LoadString(hLangLib, STR_OPTIONS, buf, BUFFER_SIZE);
  psh.pszCaption=buf;

  cbtHook=SetWindowsHookEx(WH_CBT, CBTProc, NULL, GetCurrentThreadId());

  PropertySheet(&psh);
}

void DoWindowTabView(int nView, BOOL bFirst)
{
  RECT rcRect;
  int i;

  CheckMenuItem(hMainMenu, IDM_WINDOW_TABVIEW_TOP, (nView == TAB_VIEW_TOP)?MF_CHECKED:MF_UNCHECKED);
  CheckMenuItem(hMainMenu, IDM_WINDOW_TABVIEW_BOTTOM, (nView == TAB_VIEW_BOTTOM)?MF_CHECKED:MF_UNCHECKED);
  CheckMenuItem(hMainMenu, IDM_WINDOW_TABVIEW_NONE, (nView == TAB_VIEW_NONE)?MF_CHECKED:MF_UNCHECKED);
  EnableMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_STANDART, nView?MF_ENABLED:MF_GRAYED);
  EnableMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_BUTTONS, nView?MF_ENABLED:MF_GRAYED);
  if (!bOldComctl32) EnableMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_FLATBUTTONS, nView?MF_ENABLED:MF_GRAYED);
  if (bFirst != TRUE && nView == nTabView) return;
  nTabView=nView;

  if (nTabView == TAB_VIEW_TOP)
  {
    i=GetWindowLong(hTab, GWL_STYLE) & ~TCS_BOTTOM;
    SetWindowLong(hTab, GWL_STYLE, i);
  }
  else if (nTabView == TAB_VIEW_BOTTOM)
  {
    i=GetWindowLong(hTab, GWL_STYLE);
    SetWindowLong(hTab, GWL_STYLE, i|TCS_BOTTOM);
  }
  ShowWindow(hTab, nTabView?SW_SHOW:SW_HIDE);
  GetClientRect(hMainWnd, &rcRect);
  UpdateSize(rcRect.right, rcRect.bottom);
}

void DoWindowTabType(int nType, BOOL bFirst)
{
  int i;

  CheckMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_STANDART, (nType == TAB_TYPE_STANDART)?MF_CHECKED:MF_UNCHECKED);
  CheckMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_BUTTONS, (nType == TAB_TYPE_BUTTONS)?MF_CHECKED:MF_UNCHECKED);
  CheckMenuItem(hMainMenu, IDM_WINDOW_TABTYPE_FLATBUTTONS, (nType == TAB_TYPE_FLATBUTTONS)?MF_CHECKED:MF_UNCHECKED);
  if (bFirst != TRUE && nType == nTabType) return;
  nTabType=nType;

  if (nTabType == TAB_TYPE_STANDART)
  {
    i=GetWindowLong(hTab, GWL_STYLE) & ~TCS_BUTTONS & ~TCS_FLATBUTTONS;
    SetWindowLong(hTab, GWL_STYLE, i);
    SendMessage(hTab, TCM_SETITEMSIZE, 0, (LPARAM)MAKELPARAM(TAB_WIDTH, TAB_HEIGHT - 4));
  }
  else if (nTabType == TAB_TYPE_BUTTONS)
  {
    i=GetWindowLong(hTab, GWL_STYLE) & ~TCS_FLATBUTTONS;
    SetWindowLong(hTab, GWL_STYLE, i|TCS_BUTTONS);
    SendMessage(hTab, TCM_SETITEMSIZE, 0, (LPARAM)MAKELPARAM(TAB_WIDTH, TAB_HEIGHT));
  }
  else if (nTabType == TAB_TYPE_FLATBUTTONS)
  {
    i=GetWindowLong(hTab, GWL_STYLE);
    SetWindowLong(hTab, GWL_STYLE, i|TCS_BUTTONS|TCS_FLATBUTTONS);
    SendMessage(hTab, TCM_SETITEMSIZE, 0, (LPARAM)MAKELPARAM(TAB_WIDTH, TAB_HEIGHT));
  }
}

void DoHelpAbout()
{
  API_DialogBoxA(hLangLib, MAKEINTRESOURCE(IDD_ABOUT), hMainWnd, (DLGPROC)AboutDlgProc);
}

void DoNonMenuDelLine(HWND hWnd)
{
  int nLine;
  CHARRANGE selrng;

  nLine=SendMessage(hWnd, EM_LINEFROMCHAR, nPosition, 0);
  selrng.cpMin=SendMessage(hWnd, EM_LINEINDEX, nLine, 0);
  selrng.cpMax=selrng.cpMin + SendMessage(hWnd, EM_LINELENGTH, nPosition, 0) + 1;

  SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&selrng);
  SendMessage(hWnd, EM_REPLACESEL, TRUE, (LPARAM)"");
}


//// Read/Save options

BOOL RegReadOptions()
{
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;

  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Options", 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return FALSE;

  dwSize=MAX_PATH;
  RegQueryValueEx(hKey, "LanguageModule", NULL, &dwType, (LPBYTE)szLangModule, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "WordWrap", NULL, &dwType, (LPBYTE)&bWordWrap, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "OnTop", NULL, &dwType, (LPBYTE)&bOnTop, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "StatusBar", NULL, &dwType, (LPBYTE)&bStatusBar, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "SaveTime", NULL, &dwType, (LPBYTE)&bSaveTime, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "KeepSpace", NULL, &dwType, (LPBYTE)&bKeepSpace, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "SingleOpenFile", NULL, &dwType, (LPBYTE)&bSingleOpenFile, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "SingleOpenProgram", NULL, &dwType, (LPBYTE)&bSingleOpenProgram, &dwSize);

  dwSize=sizeof(BOOL);
  RegQueryValueEx(hKey, "MDI", NULL, &dwType, (LPBYTE)&bRegMDI, &dwSize);

  dwSize=sizeof(LOGFONT);
  RegQueryValueEx(hKey, "Font", NULL, &dwType, (LPBYTE)&logfont, &dwSize);

  dwSize=sizeof(COLORREF);
  RegQueryValueEx(hKey, "TextColor", NULL, &dwType, (LPBYTE)&crFont, &dwSize);

  dwSize=sizeof(COLORREF);
  RegQueryValueEx(hKey, "BackgroundColor", NULL, &dwType, (LPBYTE)&crBackground, &dwSize);

  dwSize=sizeof(RECT);
  RegQueryValueEx(hKey, "MarginsPrint", NULL, &dwType, (LPBYTE)&psdPage.rtMargin, &dwSize);

  dwSize=sizeof(DWORD);
  RegQueryValueEx(hKey, "MarginsEdit", NULL, &dwType, (LPBYTE)&dwEditMargins, &dwSize);

  dwSize=sizeof(int);
  RegQueryValueEx(hKey, "DefaultCodepage", NULL, &dwType, (LPBYTE)&nDefaultCodePage, &dwSize);

  dwSize=sizeof(int);
  RegQueryValueEx(hKey, "RecentFiles", NULL, &dwType, (LPBYTE)&nRecentFiles, &dwSize);

  dwSize=sizeof(int);
  RegQueryValueEx(hKey, "SearchStrings", NULL, &dwType, (LPBYTE)&nSearchStrings, &dwSize);

  dwSize=sizeof(DWORD);
  RegQueryValueEx(hKey, "SearchOptions", NULL, &dwType, (LPBYTE)&ftflags, &dwSize);

  dwSize=DELIMITERS_SIZE * sizeof(wchar_t);
  RegQueryValueEx(hKey, "WordDelimiters", NULL, &dwType, (LPBYTE)wszDelimiters, &dwSize);

  dwSize=BUFFER_SIZE;
  RegQueryValueEx(hKey, "ExecuteCommand", NULL, &dwType, (LPBYTE)szCommand, &dwSize);

  dwSize=MAX_PATH;
  RegQueryValueEx(hKey, "FileTypesOpen", NULL, &dwType, (LPBYTE)szFileTypesOpen, &dwSize);

  dwSize=MAX_PATH;
  RegQueryValueEx(hKey, "FileTypesEdit", NULL, &dwType, (LPBYTE)szFileTypesEdit, &dwSize);

  dwSize=MAX_PATH;
  RegQueryValueEx(hKey, "FileTypesPrint", NULL, &dwType, (LPBYTE)szFileTypesPrint, &dwSize);

  dwSize=sizeof(DWORD);
  RegQueryValueEx(hKey, "FileTypesAssociated", NULL, &dwType, (LPBYTE)&dwFileTypesAssociated, &dwSize);

  dwSize=sizeof(DWORD);
  RegQueryValueEx(hKey, "WindowStyle", NULL, &dwType, (LPBYTE)&nMainStyle, &dwSize);

  dwSize=sizeof(RECT);
  RegQueryValueEx(hKey, "WindowPosition", NULL, &dwType, (LPBYTE)&rcMainWindow, &dwSize);

  if (bRegMDI)
  {
    dwSize=sizeof(int);
    RegQueryValueEx(hKey, "TabViewMDI", NULL, &dwType, (LPBYTE)&nTabView, &dwSize);
  
    dwSize=sizeof(int);
    RegQueryValueEx(hKey, "TabTypeMDI", NULL, &dwType, (LPBYTE)&nTabType, &dwSize);
  
    dwSize=sizeof(DWORD);
    RegQueryValueEx(hKey, "WindowStyleMDI", NULL, &dwType, (LPBYTE)&nMdiStyle, &dwSize);
  }

  RegCloseKey(hKey);
  return TRUE;
}

void RegReadSearch()
{
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;
  char *szData;

  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Search", 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return;

  if (bOldWindows)
  {
    if (RegQueryValueEx(hKey, "find0", NULL, &dwType, NULL, &dwSize) == ERROR_SUCCESS && dwSize > 0)
    {
      if (szData=(char *)API_HeapAlloc(hHeap, 0, dwSize + 1))
      {
        if (RegQueryValueEx(hKey, "find0", NULL, &dwType, (LPBYTE)szData, &dwSize) == ERROR_SUCCESS)
        {
          if (wszFind_orig=wszFind=(wchar_t *)API_HeapAlloc(hHeap, 0, dwSize * sizeof(wchar_t) + 2))
          {
            MultiByteToWideChar(CP_ACP, 0, szData, dwSize, wszFind_orig, dwSize);

            if (ftflags & FR_ESCAPESEQ)
            {
              if (wszFind=(wchar_t *)API_HeapAlloc(hHeap, 0, dwSize + 2))
              {
                EscapeStringToEscapeDataW(wszFind_orig, wszFind);
              }
            }
          }
        }
        API_HeapFree(hHeap, 0, (LPVOID)szData);
      }
    }
  }
  else
  {
    if (RegQueryValueExW(hKey, L"find0", NULL, &dwType, NULL, &dwSize) == ERROR_SUCCESS && dwSize > 0)
    {
      if (wszFind_orig=wszFind=(wchar_t *)API_HeapAlloc(hHeap, 0, dwSize + 1))
      {
        if (RegQueryValueExW(hKey, L"find0", NULL, &dwType, (LPBYTE)wszFind_orig, &dwSize) == ERROR_SUCCESS)
        {
          if (ftflags & FR_ESCAPESEQ)
          {
            if (wszFind=(wchar_t *)API_HeapAlloc(hHeap, 0, dwSize + 2))
            {
              EscapeStringToEscapeDataW(wszFind_orig, wszFind);
            }
          }
        }
      }
    }
  }
  RegCloseKey(hKey);
}

BOOL RegSaveOptions()
{
  HKEY hKey;

  nMainStyle=GetWindowLong(hMainWnd, GWL_STYLE);
  nMainStyle=((nMainStyle & WS_MAXIMIZE) || ((nMainStyle & WS_MINIMIZE) && nLastMainSize == SIZE_MAXIMIZED))?WS_MAXIMIZE:0;
  nMdiStyle=(bMdiMaximize == TRUE)?WS_MAXIMIZE:0;

  if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Options", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) != ERROR_SUCCESS)
    return FALSE;

  if (RegSetValueEx(hKey, "LanguageModule", 0, REG_SZ, (LPBYTE)szLangModule, lstrlen(szLangModule) + 1) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "WordWrap", 0, REG_DWORD, (LPBYTE)&bWordWrap, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "OnTop", 0, REG_DWORD, (LPBYTE)&bOnTop, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "StatusBar", 0, REG_DWORD, (LPBYTE)&bStatusBar, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "SaveTime", 0, REG_DWORD, (LPBYTE)&bSaveTime, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "KeepSpace", 0, REG_DWORD, (LPBYTE)&bKeepSpace, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "SingleOpenFile", 0, REG_DWORD, (LPBYTE)&bSingleOpenFile, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "SingleOpenProgram", 0, REG_DWORD, (LPBYTE)&bSingleOpenProgram, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "MDI", 0, REG_DWORD, (LPBYTE)&bRegMDI, sizeof(BOOL)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "DefaultCodepage", 0, REG_DWORD, (LPBYTE)&nDefaultCodePage, sizeof(int)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "RecentFiles", 0, REG_DWORD, (LPBYTE)&nRecentFiles, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "SearchStrings", 0, REG_DWORD, (LPBYTE)&nSearchStrings, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "SearchOptions", 0, REG_DWORD, (LPBYTE)&ftflags, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "WordDelimiters", 0, REG_BINARY, (LPBYTE)wszDelimiters, API_lstrlenW(wszDelimiters) * sizeof(wchar_t) + 2) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "ExecuteCommand", 0, REG_SZ, (LPBYTE)szCommand, lstrlen(szCommand) + 1) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "FileTypesOpen", 0, REG_SZ, (LPBYTE)szFileTypesOpen, lstrlen(szFileTypesOpen) + 1) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "FileTypesEdit", 0, REG_SZ, (LPBYTE)szFileTypesEdit, lstrlen(szFileTypesEdit) + 1) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "FileTypesPrint", 0, REG_SZ, (LPBYTE)szFileTypesPrint, lstrlen(szFileTypesPrint) + 1) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "FileTypesAssociated", 0, REG_DWORD, (LPBYTE)&dwFileTypesAssociated, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "MarginsEdit", 0, REG_DWORD, (LPBYTE)&dwEditMargins, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "MarginsPrint", 0, REG_BINARY, (LPBYTE)&psdPage.rtMargin, sizeof(RECT)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "Font", 0, REG_BINARY, (LPBYTE)&logfont, sizeof(LOGFONT)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "TextColor", 0, REG_DWORD, (LPBYTE)&crFont, sizeof(COLORREF)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "BackgroundColor", 0, REG_DWORD, (LPBYTE)&crBackground, sizeof(COLORREF)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "WindowStyle", 0, REG_DWORD, (LPBYTE)&nMainStyle, sizeof(DWORD)) != ERROR_SUCCESS)
    goto Error;

  if (RegSetValueEx(hKey, "WindowPosition", 0, REG_BINARY, (LPBYTE)&rcMainWindow, sizeof(RECT)) != ERROR_SUCCESS)
    goto Error;

  if (bMDI)
  {
    if (RegSetValueEx(hKey, "TabViewMDI", 0, REG_DWORD, (LPBYTE)&nTabView, sizeof(DWORD)) != ERROR_SUCCESS)
      goto Error;

    if (RegSetValueEx(hKey, "TabTypeMDI", 0, REG_DWORD, (LPBYTE)&nTabType, sizeof(DWORD)) != ERROR_SUCCESS)
      goto Error;

    if (RegSetValueEx(hKey, "WindowStyleMDI", 0, REG_DWORD, (LPBYTE)&nMdiStyle, sizeof(DWORD)) != ERROR_SUCCESS)
      goto Error;
  }

  RegCloseKey(hKey);
  return TRUE;

  Error:
  RegCloseKey(hKey);
  return FALSE;
}


//// Open/Save document

int OpenDocument(char *szFile, DWORD Flags, int nCodePage, BOOL bBOM)
{
  HANDLE hFile;
  STREAMDATA sd;
  EDITSTREAM es;
  HWND hWndFriend;
  char *pFileName;
  int i;

  //Check that file exists
  if (!FileExists(szFile))
  {
    API_LoadString(hLangLib, MSG_FILE_DOES_NOT_EXIST, buf, BUFFER_SIZE);
    wsprintf(buf2, buf, szFile);
    if (MessageBox(hMainWnd, buf2, APP_SHORT_TITLE, MB_OKCANCEL|MB_ICONEXCLAMATION) == IDCANCEL)
      return EDT_CANCEL;
  }

  //Get full path
  GetFullPathName(szFile, BUFFER_SIZE, buf, &pFileName);
  lstrcpy(szFile, buf);
  pFileName=szFile + (pFileName - buf);

  if (bMDI && !bMdiReopen && bSingleOpenFile)
  {
    if (hWndFriend=FindWindowEx(hMdiClient, NULL, APP_MDI_NAME, szFile))
    {
      SetFocus(hWndFriend);
      return 0;
    }
  }

  //Autodetect code page
  if ((i=AutodetectCodePage(szFile, DETECTBINARY_SIZE, Flags, &nCodePage, &bBOM)) < 0)
  {
    if (i == EDT_BINARY)
    {
      API_LoadString(hLangLib, MSG_ERROR_BINARY, buf, BUFFER_SIZE);
      wsprintf(buf2, buf, szFile);
      if (MessageBox(hMainWnd, buf2, APP_SHORT_TITLE, MB_OKCANCEL|MB_ICONEXCLAMATION|MB_DEFBUTTON2) == IDCANCEL)
        return EDT_CANCEL;
    }
    else return i;
  }

  //Open file
  if ((hFile=API_CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return EDT_OPEN;

  //Save position of the previous file before load new document
  if (nRecentFiles && *szCurrentFile)
  {
    RecentFilesZero();
    RecentFilesRead();
    RecentFilesUpdate(szCurrentFile, nPosition, MAKELONG(nCurrentCodePage, bCurrentBOM));
    RecentFilesSave();
  }

  if (bBOM)
  {
    if (nCodePage == CP_UNICODE_UCS2_LE)
      SetFilePointer(hFile, 2, NULL, FILE_BEGIN);
    else if (nCodePage == CP_UNICODE_UCS2_BE)
      SetFilePointer(hFile, 2, NULL, FILE_BEGIN);
    else if (nCodePage == CP_UNICODE_UTF8)
      SetFilePointer(hFile, 3, NULL, FILE_BEGIN);
  }

  if (bMDI && !bMdiReopen && (!hWndMdiActive || bModified || *szCurrentFile))
  {
    if (hWndMdiActive) SendMessage(hMdiClient, WM_MDIGETACTIVE, 0, (LPARAM)&bMdiMaximize);
    CreateMDIWindow(APP_MDI_NAME, "", (bMdiMaximize == TRUE)?WS_MAXIMIZE:0, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, hMdiClient, hInstance, 0);
  }
  SetUnixNewLine(FALSE, FALSE);
  bNewLineDetected=FALSE;
  HideCaret(NULL);

  //Load text
  sd.hFile=hFile;
  sd.nCodePage=nCodePage;
  sd.nBytes=-1;
  sd.nBytesRead=0;
  es.dwCookie=(DWORD)&sd;
  es.pfnCallback=InputStreamCallbackW;
  SendMessage(hWndEdit, EM_STREAMIN, SF_TEXT|SF_UNICODE, (LPARAM)&es);

  ShowCaret(NULL);
  SetModify(FALSE, FALSE);
  CloseHandle(hFile);

  //Update titles and icons
  if (szCurrentFile != szFile)
    UpdateTitle(szFile);

  //Read position of the new document
  if (szCurrentFile != szFile || nCurrentCodePage != nCodePage)
  {
    if (nRecentFiles)
    {
      RecentFilesZero();
      RecentFilesRead();
      RecentFilesUpdate(szFile, -1, MAKELONG(nCodePage, bBOM));
      RecentFilesSave();
    }
    if (szCurrentFile != szFile)
    {
      if (nRecentFiles)
        RecentFilesMenu();
      lstrcpy(szCurrentFile, szFile);
    }
  }
  if (nRecentFiles)
    SendMessage(hWndEdit, EM_SETSEL, szRecentPositions[0], szRecentPositions[0]);

  //Update codepage
  SetCodePage(MAKELONG(nCodePage, bBOM), FALSE);

  //Print if "/p" option used in command line
  if (bGlobalPrint)
  {
    DoFilePrint();
    bGlobalPrint=FALSE;

    if (!bMDI)
    {
      SendMessage(hMainWnd, WM_CLOSE, 0, 0);
    }
    else
    {
      SendMessage(hMdiClient, WM_MDIDESTROY, (WPARAM)hWndMdiActive, 0);
      if (!bSingleOpenProgram && !hWndMdiActive) SendMessage(hMainWnd, WM_CLOSE, 0, 0);
    }
  }
  return 0;
}

DWORD CALLBACK InputStreamCallbackW(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
  static int nBytesConverted;
  STREAMDATA *lpData=(STREAMDATA *)dwCookie;

  if (nBytesConverted == 0)
  {
    if (lpData->nBytes != -1 && (lpData->nBytesRead+=cb) > lpData->nBytes)
      return 0xFFFFFFFF;

    if (!API_ReadFile(lpData->hFile, pbBuff, (DWORD)cb, (DWORD *)pcb, NULL))
      return 0xFFFFFFFF;

    if (*pcb == 0)
      return 0xFFFFFFFF;
  }

  if (lpData->nCodePage == CP_UNICODE_UCS2_LE)
  {
  }
  else if (lpData->nCodePage == CP_UNICODE_UCS2_BE)
  {
    ChangeByteOrder(pbBuff, *pcb);
  }
  else
  {
    if (nBytesConverted == 0)
    {
      int nBytesRead=*pcb;

      if (lpData->nCodePage == CP_UNICODE_UTF8)
        nBytesConverted=UTF8toUTF16((char *)pbBuff, (unsigned int *)&nBytesRead, (wchar_t *)pcTranslateBuffer, TRANSLATE_BUFFER_SIZE / sizeof(wchar_t)) * sizeof(wchar_t);
      else
        nBytesConverted=MultiByteToWideChar(lpData->nCodePage, 0, (char *)pbBuff, *pcb, (wchar_t *)pcTranslateBuffer, TRANSLATE_BUFFER_SIZE / sizeof(wchar_t)) * sizeof(wchar_t);

      if (lpData->nCodePage == CP_UNICODE_UTF8 && nBytesRead != *pcb)
        SetFilePointer(lpData->hFile, nBytesRead - *pcb, NULL, FILE_CURRENT);

      if (nBytesConverted <= cb)
      {
        memcpy(pbBuff, pcTranslateBuffer, nBytesConverted);
        *pcb=nBytesConverted;
        nBytesConverted=0;
      }
      else
      {
        memcpy(pbBuff, pcTranslateBuffer, cb);
        *pcb=cb;
        nBytesConverted-=cb;
      }
    }
    else
    {
      memcpy(pbBuff, pcTranslateBuffer + cb, nBytesConverted);
      *pcb=nBytesConverted;
      nBytesConverted=0;
    }
  }

  if (!bNewLineDetected)
  {
    wchar_t *wpbBuff=(wchar_t *)pbBuff;
    int nBytes=*pcb / sizeof(wchar_t);
    int i;

    for (i=0; i < nBytes; ++i)
    {
      if (wpbBuff[i] == '\n')
      {
        if (i == 0 || wpbBuff[i - 1] != '\r')
          SetUnixNewLine(TRUE, FALSE);
        bNewLineDetected=TRUE;
        break;
      }
      else if (wpbBuff[i] == '\r')
      {
        bNewLineDetected=TRUE;
      }
    }
  }
  return 0;
}

BOOL SaveDocument(char *szFile, int nCodePage, BOOL bBOM)
{
  WIN32_FIND_DATA w32fdInfo;
  HANDLE hFile;
  STREAMDATA sd;
  EDITSTREAM es;
  DWORD dwAttr;
  unsigned int nBytesWritten;
  int nError;

  dwAttr=GetFileAttributes(szFile);
  if (dwAttr != 0xFFFFFFFF)
  {
    if ((dwAttr & FILE_ATTRIBUTE_READONLY) || (dwAttr & FILE_ATTRIBUTE_HIDDEN) || (dwAttr & FILE_ATTRIBUTE_SYSTEM))
      SetFileAttributes(szFile, dwAttr & (!FILE_ATTRIBUTE_READONLY) & (!FILE_ATTRIBUTE_HIDDEN) & (!FILE_ATTRIBUTE_SYSTEM));
    if (bSaveTime)
    {
      if ((hFile=FindFirstFile(szFile, &w32fdInfo)) != INVALID_HANDLE_VALUE)
        FindClose(hFile);
    }
  }

  hFile=API_CreateFile(szFile, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL);

  if (hFile == INVALID_HANDLE_VALUE)
  {
    if (dwAttr != 0xFFFFFFFF) SetFileAttributes(szFile, dwAttr);
    return FALSE;
  }

  if (bBOM)
  {
    if (nCodePage == CP_UNICODE_UCS2_LE)
      nError=API_WriteFile(hFile, "\xFF\xFE", 2, (DWORD *)&nBytesWritten, NULL);
    else if (nCodePage == CP_UNICODE_UCS2_BE)
      nError=API_WriteFile(hFile, "\xFE\xFF", 2, (DWORD *)&nBytesWritten, NULL);
    else if (nCodePage == CP_UNICODE_UTF8)
      nError=API_WriteFile(hFile, "\xEF\xBB\xBF", 3, (DWORD *)&nBytesWritten, NULL);

    if (!nError)
    {
      if (dwAttr != 0xFFFFFFFF) SetFileAttributes(szFile, dwAttr);
      return FALSE;
    }
  }

  sd.hFile=hFile;
  sd.nCodePage=nCodePage;
  es.dwCookie=(DWORD)&sd;
  es.pfnCallback=OutputStreamCallbackW;
  SendMessage(hWndEdit, EM_STREAMOUT, SF_TEXT|SF_UNICODE, (LPARAM)&es);

  SetModify(FALSE, FALSE);
  CloseHandle(hFile);

  //Change back file attributes
  if (dwAttr != 0xFFFFFFFF)
  {
    SetFileAttributes(szFile, dwAttr);
    if (bSaveTime)
    {
      if ((hFile=API_CreateFile(szFile, FILE_WRITE_ATTRIBUTES, 0, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL)) != INVALID_HANDLE_VALUE)
      {
        SetFileTime(hFile, NULL, NULL, &w32fdInfo.ftLastWriteTime);
        CloseHandle(hFile);
      }
    }
  }

  //Update titles and icons
  if (szCurrentFile != szFile)
    UpdateTitle(szFile);

  //Save position of the document
  if (szCurrentFile != szFile || nCurrentCodePage != nCodePage)
  {
    if (nRecentFiles)
    {
      RecentFilesZero();
      RecentFilesRead();
      RecentFilesUpdate(szFile, nPosition, MAKELONG(nCodePage, bBOM));
      RecentFilesSave();
    }
    if (szCurrentFile != szFile)
    {
      if (nRecentFiles)
        RecentFilesMenu();
      lstrcpy(szCurrentFile, szFile);
    }
  }

  //Update codepage
  SetCodePage(MAKELONG(nCodePage, bBOM), FALSE);

  return TRUE;
}

DWORD CALLBACK OutputStreamCallbackW(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
  unsigned int nBytesConverted;
  unsigned int nBytesToWrite;
  unsigned int nBytesWritten;
  unsigned int nBytesRead;
  unsigned char *pDataToWrite=pbBuff;
  STREAMDATA *lpData=(STREAMDATA *)dwCookie;

  if (bUnixNewLine)
  {
    nBytesRead=cb/2;
    nBytesToWrite=TranslateNewLinesToUnixW((wchar_t *)pbBuff, &nBytesRead) * sizeof(wchar_t);
  }
  else nBytesToWrite=cb;

  if (lpData->nCodePage == CP_UNICODE_UCS2_LE)
  {
  }
  else if (lpData->nCodePage == CP_UNICODE_UCS2_BE)
  {
    ChangeByteOrder(pbBuff, nBytesToWrite);
  }
  else
  {
    if (lpData->nCodePage == CP_UNICODE_UTF8)
      nBytesConverted=UTF16toUTF8((wchar_t *)pbBuff, nBytesToWrite/2, (char *)pcTranslateBuffer, TRANSLATE_BUFFER_SIZE);
    else
      nBytesConverted=WideCharToMultiByte(lpData->nCodePage, 0, (wchar_t *)pbBuff, nBytesToWrite/2, (char *)pcTranslateBuffer, TRANSLATE_BUFFER_SIZE, NULL, NULL);

    pDataToWrite=pcTranslateBuffer;
    nBytesToWrite=nBytesConverted;
  }

  if (!API_WriteFile(lpData->hFile, pDataToWrite, nBytesToWrite, (DWORD *)&nBytesWritten, NULL))
    return 0xFFFFFFFF;

  *pcb=cb;

  return 0;
}


//// Code pages

unsigned int CALLBACK CodePageDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndCP;
  static HWND hWndPreview;
  static HWND hWndAutodetect;
  static HWND hDlgEdit;
  static RECT rcDlg;
  static RECT rcCP;
  static RECT rcPreview;
  static RECT rcAutodetect;
  static char szFile[MAX_PATH];
  static int nCodePage;
  static BOOL bBOM;
  int i;

  if (uMsg == WM_INITDIALOG)
  {
    if (bMDI) hDlgEdit=FindWindowEx(GetParent(hDlg), NULL, "Edit", NULL);
    hWndCP=GetDlgItem(hDlg, IDC_OFN_CODEPAGE);
    hWndPreview=GetDlgItem(hDlg, IDC_OFN_PREVIEW);
    hWndAutodetect=GetDlgItem(hDlg, IDC_OFN_AUTODETECT);

    GetWindowRect(hDlg, &rcDlg);
    GetWindowRect(hWndCP, &rcCP);
    GetWindowRect(hWndPreview, &rcPreview);
    rcDlg.right-=rcDlg.left;
    rcDlg.bottom-=rcDlg.top;
    rcCP.right-=rcCP.left;
    rcCP.bottom-=rcCP.top;
    rcPreview.right-=rcPreview.left;
    rcPreview.bottom-=rcPreview.top;

    nCodePage=nCurrentCodePage;
    FillComboboxCodepage(hWndCP);
    SelectComboboxCodepage(hWndCP, nCodePage);

    if (bSaveDlg)
    {
      SetWindowPos(hDlg, 0, 0, 0, rcDlg.right, rcDlg.bottom - ((rcDlg.bottom + rcDlg.top) - rcPreview.top), SWP_NOMOVE|SWP_NOZORDER);
      SendMessage(hWndAutodetect, WM_SETTEXT, 0, (LPARAM)"&BOM");

      if (nCodePage == CP_UNICODE_UCS2_LE || nCodePage == CP_UNICODE_UCS2_BE || nCodePage == CP_UNICODE_UTF8)
      {
        SendMessage(hWndAutodetect, BM_SETCHECK, (WPARAM)bCurrentBOM, 0);
        EnableWindow(hWndAutodetect, TRUE);
      }
      else
      {
        SendMessage(hWndAutodetect, BM_SETCHECK, (WPARAM)FALSE, 0);
        EnableWindow(hWndAutodetect, FALSE);
      }
    }
    else
    {
      API_LoadString(hLangLib, STR_AUTODETECT, buf, BUFFER_SIZE);
      SendMessage(hWndAutodetect, WM_SETTEXT, 0, (LPARAM)buf);
      SendMessage(hWndAutodetect, BM_SETCHECK, (WPARAM)bAutodetect, 0);
      EnableWindow(hWndCP, !bAutodetect);
    }

    OldPreviewProc=(WNDPROC)GetWindowLong(hWndPreview, GWL_WNDPROC);
    SetWindowLong(hWndPreview, GWL_WNDPROC, (LONG)NewPreviewProc);
  }
  else if (uMsg == WM_SIZE)
  {
    RECT rcTmp;

    GetWindowRect(hDlg, &rcTmp);
    GetWindowRect(hWndAutodetect, &rcAutodetect);
    rcAutodetect.left-=rcTmp.left;
    rcAutodetect.top-=rcTmp.top;

    i=LOWORD(lParam) - rcDlg.right;
    rcDlg.right=LOWORD(lParam);
    rcDlg.bottom=HIWORD(lParam);
    rcCP.right+=i;
    rcPreview.right+=i;
    rcAutodetect.left+=i;

    SetWindowPos(hWndCP, 0, 0, 0, rcCP.right, rcCP.bottom, SWP_NOMOVE|SWP_NOZORDER);
    SetWindowPos(hWndPreview, 0, 0, 0, rcPreview.right, rcPreview.bottom, SWP_NOMOVE|SWP_NOZORDER);
    SetWindowPos(hWndAutodetect, 0, rcAutodetect.left, rcAutodetect.top, 0, 0, SWP_NOSIZE|SWP_NOZORDER);
    return TRUE;
  }
  else if (uMsg == WM_CONTEXTMENU)
  {
    if ((HWND)wParam == hWndPreview)
    {
      RECT rcRect;
      POINT pt;
      CHARRANGE cr;

      if (lParam == ((UINT)-1))
      {
        GetWindowRect((HWND)wParam, &rcRect);
        pt.x=rcRect.left;
        pt.y=rcRect.top;
      }
      else
      {
        pt.x=LOWORD(lParam);
        pt.y=HIWORD(lParam);
      }
      SendMessage((HWND)wParam, EM_EXGETSEL, 0, (LPARAM)&cr);
      EnableMenuItem(hPopupView, IDM_EDIT_COPY, (cr.cpMin == cr.cpMax)?MF_GRAYED:MF_ENABLED);
      i=TrackPopupMenu(hPopupView, TPM_RETURNCMD|TPM_NONOTIFY, pt.x, pt.y, 0, hDlg, NULL);

      if (i == IDM_EDIT_COPY)
        SendMessage((HWND)wParam, WM_COPY, 0, 0);
      else if (i == IDM_EDIT_SELECTALL)
        SendMessage((HWND)wParam, EM_SETSEL, 0, -1);
    }
    return TRUE;
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (((LPNMHDR)lParam)->code == CDN_FILEOK)
    {
      char ch='';
      wchar_t wch;

      nCodePage=GetComboboxCodepage(hWndCP);

      if (!bSaveDlg && bAutodetect)
      {
        dwOfnFlags=ADT_BINARY_ERROR|ADT_REG_CODEPAGE;
        nOfnCodePage=0;
        bOfnBOM=FALSE;
      }
      else if (nCodePage == CP_UNICODE_UCS2_LE || nCodePage == CP_UNICODE_UCS2_BE || nCodePage == CP_UNICODE_UTF8)
      {
        dwOfnFlags=0;
        nOfnCodePage=nCodePage;
        if (!bSaveDlg)
          bOfnBOM=bBOM;
        else
          bOfnBOM=(BOOL)SendMessage(hWndAutodetect, BM_GETCHECK, 0, 0);
      }
      else if (MultiByteToWideChar(nCodePage, 0, &ch, 1, &wch, 1))
      {
        dwOfnFlags=0;
        nOfnCodePage=nCodePage;
        bOfnBOM=FALSE;
      }
      else
      {
        API_LoadString(hLangLib, MSG_CP_UNIMPLEMENTED, buf, BUFFER_SIZE);
        MessageBox(hDlg, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
        SetWindowLong(hDlg, DWL_MSGRESULT, 1);
      }
    }
    else if (((LPNMHDR)lParam)->code == CDN_SELCHANGE)
    {
      if (!bSaveDlg)
      {
        if (bMDI)
        {
          if (GetWindowTextLength(hDlgEdit) > FILELIST_BUFFER_SIZE)
          {
            API_LoadString(hLangLib, MSG_LONG_FILELIST, buf, BUFFER_SIZE);
            MessageBox(hDlg, buf, APP_SHORT_TITLE, MB_OK|MB_ICONEXCLAMATION);
          }
        }

        SendMessage(((OFNOTIFY*)lParam)->hdr.hwndFrom, CDM_GETFILEPATH, (LPARAM)MAX_PATH, (WPARAM)szFile);

        if (FilePreview(hWndPreview, szFile, PREVIEW_SIZE, bAutodetect?(ADT_BINARY_ERROR|ADT_REG_CODEPAGE):ADT_DETECT_BOM, &nCodePage, &bBOM) < 0)
        {
          EnableWindow(hWndPreview, FALSE);
          if (bAutodetect) SelectComboboxCodepage(hWndCP, nDefaultCodePage);
          SendMessage(hWndPreview, WM_SETTEXT, 0, (LPARAM)"");
          return TRUE;
        }
        EnableWindow(hWndPreview, TRUE);
        if (bAutodetect) SelectComboboxCodepage(hWndCP, nCodePage);
      }
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_OFN_CODEPAGE && HIWORD(wParam) == CBN_SELCHANGE)
    {
      nCodePage=GetComboboxCodepage(hWndCP);

      if (bSaveDlg)
      {
        if (nCodePage == CP_UNICODE_UCS2_LE || nCodePage == CP_UNICODE_UCS2_BE || nCodePage == CP_UNICODE_UTF8)
        {
          EnableWindow(hWndAutodetect, TRUE);
          SendMessage(hWndAutodetect, BM_SETCHECK, (WPARAM)TRUE, 0);
        }
        else
        {
          EnableWindow(hWndAutodetect, FALSE);
          SendMessage(hWndAutodetect, BM_SETCHECK, (WPARAM)FALSE, 0);
        }
      }
      else
      {
        if (FilePreview(hWndPreview, szFile, PREVIEW_SIZE, ADT_DETECT_BOM, &nCodePage, &bBOM) < 0)
          SendMessage(hWndPreview, WM_SETTEXT, 0, (LPARAM)"");
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OFN_AUTODETECT)
    {
      if (!bSaveDlg)
      {
        bAutodetect=(BOOL)SendMessage(hWndAutodetect, BM_GETCHECK, 0, 0);
        EnableWindow(hWndCP, !bAutodetect);

        if (FilePreview(hWndPreview, szFile, PREVIEW_SIZE, bAutodetect?(ADT_BINARY_ERROR|ADT_REG_CODEPAGE):ADT_DETECT_BOM, &nCodePage, &bBOM) < 0)
        {
          EnableWindow(hWndPreview, FALSE);
          if (bAutodetect) SelectComboboxCodepage(hWndCP, nDefaultCodePage);
          SendMessage(hWndPreview, WM_SETTEXT, 0, (LPARAM)"");
          return TRUE;
        }
        EnableWindow(hWndPreview, TRUE);
        if (bAutodetect) SelectComboboxCodepage(hWndCP, nCodePage);
      }
      return TRUE;
    }
  }
  return FALSE;
}

LRESULT CALLBACK NewPreviewProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_KEYDOWN)
  {
    if (wParam == VK_ESCAPE)
    {
      PostMessage(GetParent(hWnd), WM_KEYDOWN, VK_ESCAPE, 0);
      return TRUE;
    }
  }
  return CallWindowProc(OldPreviewProc, hWnd, uMsg, wParam, lParam);
}

BOOL RegEnumCodePages(int *codepages, int size)
{
  HKEY hKey;
  DWORD dwSizeValue;
  DWORD dwSizeString;
  DWORD dwType;
  DWORD dwIndex=0;
  int i;

  memset(codepages, 0, size);

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Control\\Nls\\CodePage", 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return FALSE;

  while(1)
  {
    dwSizeValue=BUFFER_SIZE;
    dwSizeString=BUFFER_SIZE;
    buf2[0]='\0';

    if (RegEnumValue(hKey, dwIndex++, buf, &dwSizeValue, NULL, &dwType, (LPBYTE)buf2, &dwSizeString) != ERROR_SUCCESS)
      break;

    if (*buf2 != '\0' && (i=xatoi(buf)) != 0 && i < size)
      codepages[i]=TRUE;
  }
  RegCloseKey(hKey);

  return TRUE;
}

void FillComboboxCodepage(HWND hWnd)
{
  CPINFOEX CPInfoEx;
  int nCount;
  int i;

  if (!bEnumCodePages)
  {
    RegEnumCodePages(codepages, sizeof(codepages));
    if (!bOldWindows) codepages[65000]=TRUE;
    bEnumCodePages=TRUE;
  }

  if (!bRecodeDlg) SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)STR_SYSTEM_ANSI);
  SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)STR_SYSTEM_OEM);
  if (!bRecodeDlg) SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)STR_UNICODE_UCS2_LE);
  if (!bRecodeDlg) SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)STR_UNICODE_UCS2_BE);
  SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)STR_UNICODE_UTF8);
  nCount=SendMessage(hWnd, CB_GETCOUNT, 0, 0);

  for (i=0; i <= 65535; ++i)
  {
    if (codepages[i])
    {
      if (GetCPInfoExPtr && (*GetCPInfoExPtr)(i, 0, &CPInfoEx) && i == xatoi(CPInfoEx.CodePageName))
      {
        SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)CPInfoEx.CodePageName);
      }
      else
      {
        wsprintf(buf, "%d", i);
        SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)buf);
      }
      codepages[i]=nCount++;
    }
  }
}

int GetComboboxCodepage(HWND hWnd)
{
  int nCodePage;

  SendMessage(hWnd, CB_GETLBTEXT, (WPARAM)SendMessage(hWnd, CB_GETCURSEL, 0, 0), (LPARAM)buf);

  if (!lstrcmp(buf, STR_SYSTEM_ANSI))
    nCodePage=nAnsiCodePage;
  else if (!lstrcmp(buf, STR_SYSTEM_OEM))
    nCodePage=nOemCodePage;
  else
    nCodePage=xatoi(buf);

  return nCodePage;
}

int SelectComboboxCodepage(HWND hWnd, int nCodepage)
{
  int nSelectIndex;

  if (nCodepage == nAnsiCodePage)
    nSelectIndex=SendMessage(hWnd, CB_FINDSTRING, 0, (LPARAM)STR_SYSTEM_ANSI);
  else if (nCodepage == nOemCodePage)
    nSelectIndex=SendMessage(hWnd, CB_FINDSTRING, 0, (LPARAM)STR_SYSTEM_OEM);
  else if (nCodepage == CP_UNICODE_UCS2_LE)
    nSelectIndex=SendMessage(hWnd, CB_FINDSTRING, 0, (LPARAM)STR_UNICODE_UCS2_LE);
  else if (nCodepage == CP_UNICODE_UCS2_BE)
    nSelectIndex=SendMessage(hWnd, CB_FINDSTRING, 0, (LPARAM)STR_UNICODE_UCS2_BE);
  else if (nCodepage == CP_UNICODE_UTF8)
    nSelectIndex=SendMessage(hWnd, CB_FINDSTRING, 0, (LPARAM)STR_UNICODE_UTF8);
  else
    nSelectIndex=codepages[nCodepage];

  if (nSelectIndex == CB_ERR) nSelectIndex=0;
  SendMessage(hWnd, CB_SETCURSEL, (WPARAM)nSelectIndex, 0);

  return nSelectIndex;
}

int FilePreview(HWND hWnd, char *pFile, int nPreviewBytes, DWORD Flags, int *nCodePage, BOOL *bBOM)
{
  HANDLE hFile;
  STREAMDATA sd;
  EDITSTREAM es;
  int i;

  if (IsFile(pFile) != ERROR_SUCCESS)
    return EDT_OPEN;

  if ((i=AutodetectCodePage(pFile, DETECTBINARY_SIZE, Flags, nCodePage, bBOM)) < 0)
    return i;

  if ((hFile=API_CreateFile(pFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
    return EDT_OPEN;

  if (*bBOM)
  {
    if (*nCodePage == CP_UNICODE_UCS2_LE)
      SetFilePointer(hFile, 2, NULL, FILE_BEGIN);
    else if (*nCodePage == CP_UNICODE_UCS2_BE)
      SetFilePointer(hFile, 2, NULL, FILE_BEGIN);
    else if (*nCodePage == CP_UNICODE_UTF8)
      SetFilePointer(hFile, 3, NULL, FILE_BEGIN);
  }

  sd.hFile=hFile;
  sd.nCodePage=*nCodePage;
  sd.nBytes=nPreviewBytes;
  sd.nBytesRead=0;
  es.dwCookie=(DWORD)&sd;
  es.pfnCallback=InputStreamCallbackW;
  SendMessage(hWnd, EM_STREAMIN, SF_TEXT|SF_UNICODE, (LPARAM)&es);

  CloseHandle(hFile);
  return 0;
}

int AutodetectCodePage(char *pFile, DWORD dwBytesToCheck, DWORD Flags, int *nCodePage, BOOL *bBOM)
{
  HANDLE hFile;
  DWORD dwBytesRead;
  DWORD dwCount;
  unsigned char *pBuffer;
  int nUTF=0;
  int i;

  if (!dwBytesToCheck || !Flags) return 0;

  if (Flags & ADT_REG_CODEPAGE)
  {
    if (nRecentFiles)
    {
      RecentFilesZero();
      RecentFilesRead();
    }
    if (nRecentFiles && RecentFilesGet(pFile, 0, &i) && i)
    {
      if (nCodePage) *nCodePage=LOWORD(i);
      if (bBOM) *bBOM=HIWORD(i);
      Flags&=~ADT_DETECT_CODEPAGE;
      Flags&=~ADT_DETECT_BOM;
    }
    else
    {
      if (nCodePage) *nCodePage=nDefaultCodePage;
      if (bBOM) *bBOM=bDefaultBOM;
      Flags&=~ADT_REG_CODEPAGE;
      Flags|=ADT_DETECT_CODEPAGE;
      Flags|=ADT_DETECT_BOM;
    }
  }

  if (Flags & ADT_BINARY_ERROR || Flags & ADT_DETECT_CODEPAGE || Flags & ADT_DETECT_BOM)
  {
    if ((hFile=API_CreateFile(pFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE)
      return EDT_OPEN;

    if (!(pBuffer=(unsigned char *)API_HeapAlloc(hHeap, 0, dwBytesToCheck + 1)))
    {
      CloseHandle(hFile);
      return EDT_ALLOC;
    }

    if (!API_ReadFile(hFile, pBuffer, dwBytesToCheck, &dwBytesRead, NULL))
    {
      API_HeapFree(hHeap, 0, (LPVOID)pBuffer);
      CloseHandle(hFile);
      return EDT_READ;
    }
    CloseHandle(hFile);
  }

  if (Flags & ADT_BINARY_ERROR || Flags & ADT_DETECT_CODEPAGE)
  {
    if (dwBytesRead)
    {
      for (dwCount=0, --dwBytesRead; dwCount < dwBytesRead; ++dwCount)
      {
        if (Flags & ADT_BINARY_ERROR)
        {
          if (pBuffer[dwCount] == 0x00 && pBuffer[dwCount + 1] == 0x00)
          {
            API_HeapFree(hHeap, 0, (LPVOID)pBuffer);
            return EDT_BINARY;
          }
        }
        if (Flags & ADT_DETECT_CODEPAGE)
        {
          if (nUTF == 0 && dwCount % 2 == 0)
          {
            if (pBuffer[dwCount] != 0x00 && pBuffer[dwCount + 1] == 0x00)
              nUTF=CP_UNICODE_UCS2_LE;
            else if (pBuffer[dwCount] == 0x00 && pBuffer[dwCount + 1] != 0x00)
              nUTF=CP_UNICODE_UCS2_BE;
          }
        }
      }
    }
  }

  if (Flags & ADT_DETECT_CODEPAGE || Flags & ADT_DETECT_BOM)
  {
    if (pBuffer[0] == 0xFF && pBuffer[1] == 0xFE)
    {
      if (Flags & ADT_DETECT_CODEPAGE) *nCodePage=CP_UNICODE_UCS2_LE;
      if (Flags & ADT_DETECT_BOM) *bBOM=TRUE;
    }
    else if (pBuffer[0] == 0xFE && pBuffer[1] == 0xFF)
    {
      if (Flags & ADT_DETECT_CODEPAGE) *nCodePage=CP_UNICODE_UCS2_BE;
      if (Flags & ADT_DETECT_BOM) *bBOM=TRUE;
    }
    else if (pBuffer[0] == 0xEF && pBuffer[1] == 0xBB && pBuffer[2] == 0xBF)
    {
      if (Flags & ADT_DETECT_CODEPAGE) *nCodePage=CP_UNICODE_UTF8;
      if (Flags & ADT_DETECT_BOM) *bBOM=TRUE;
    }
    else if (nUTF)
    {
      *nCodePage=nUTF;
      if (Flags & ADT_DETECT_BOM) *bBOM=FALSE;
    }
    else
    {
      if (Flags & ADT_DETECT_CODEPAGE) *nCodePage=nDefaultCodePage;
      if (Flags & ADT_DETECT_BOM) *bBOM=FALSE;
    }
  }

  if (Flags & ADT_BINARY_ERROR || Flags & ADT_DETECT_CODEPAGE || Flags & ADT_DETECT_BOM)
  {
    API_HeapFree(hHeap, 0, (LPVOID)pBuffer);
  }
  return 0;
}

unsigned int UTF8toUTF16(const char *buf, unsigned int *len, wchar_t *tbuf, unsigned int tlen)
{
  const unsigned char *ubuf=(const unsigned char *)buf;
  unsigned int i;
  unsigned int ti;

  for (i=0, ti=0; i < *len && ti < tlen && ubuf[i]; ++ti)
  {
    if (ubuf[i] < 0x80)
    {
      tbuf[ti]=ubuf[i++];
    }
    else if (ubuf[i] < 0x80 + 0x40 + 0x20)
    {
      tbuf[ti]=(ubuf[i++] & 0x1F) << 6;
      if (i >= *len || !ubuf[i]) {*len=i - 1; return ti;}
      tbuf[ti]=tbuf[ti] + (ubuf[i++] & 0x7F);
    }
    else
    {
      tbuf[ti]=(ubuf[i++] & 0xF) << 12;
      if (i >= *len || !ubuf[i]) {*len=i - 1; return ti;}
      tbuf[ti]=tbuf[ti] + ((ubuf[i++] & 0x7F) << 6);
      if (i >= *len || !ubuf[i]) {*len=i - 2; return ti;}
      tbuf[ti]=tbuf[ti] + (ubuf[i++] & 0x7F);
    }
  }
  if (i < *len && ti < tlen)
    tbuf[ti++]='\0';
  *len=i;
  return ti;
}

unsigned int UTF16toUTF8(const wchar_t *tbuf, unsigned int tlen, char *buf, unsigned int len)
{
  unsigned char *ubuf=(unsigned char *)buf;
  unsigned int i;
  unsigned int ti;

  for (i=0, ti=0; ti < tlen && tbuf[ti]; ++ti)
  {
    if (tbuf[ti] < 0x80)
    {
      if (i + 1 > len) return i;
      ubuf[i++]=(unsigned char)tbuf[ti];
    }
    else if (tbuf[ti] < 0x800)
    {
      if (i + 2 > len) return i;
      ubuf[i++]=0xC0|((tbuf[ti] >> 6) & 0x1F);
      ubuf[i++]=0x80|(tbuf[ti] & 0x3F);
    }
    else
    {
      if (i + 3 > len) return i;
      ubuf[i++]=0xE0|((tbuf[ti] >> 12) & 0xF);
      ubuf[i++]=0x80|((tbuf[ti] >> 6) & 0x3F);
      ubuf[i++]=0x80|(tbuf[ti] & 0x3F);
    }
  }
  if (ti < tlen)
    ubuf[i++]='\0';
  return i;
}

void ChangeByteOrder(unsigned char *buf, unsigned int len)
{
  unsigned char ch;
  unsigned int a;
  unsigned int b;

  for (a=0, b=1; a < len; a+=2, b+=2)
  {
    ch=buf[a];
    buf[a]=buf[b];
    buf[b]=ch;
  }
}

unsigned int TranslateNewLinesToUnixW(wchar_t *tbuf, unsigned int *tlen)
{
  unsigned int a;
  unsigned int b;

  for (a=0, b=0; a < *tlen && tbuf[a]; ++a, ++b)
  {
    if (tbuf[a] == '\r' && tbuf[a + 1] == '\n') ++a;

    tbuf[b]=tbuf[a];
  }
  *tlen=a;
  return b;
}


//// Find/Replace

BOOL CALLBACK FindAndReplaceDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndFind;
  static HWND hWndReplace;
  static HWND hWndMatchCase;
  static HWND hWndWholeWord;
  static HWND hWndEscapeSeq;
  static HWND hWndForward;
  static HWND hWndBackward;
  static HWND hWndBeginning;
  static HWND hWndInSelection;
  static HWND hWndFindNextButton;
  static HWND hWndReplaceButton;
  static HWND hWndReplaceAllButton;
  static HWND hFocus;
  wchar_t *wszData;
  BOOL bReplace=FALSE;
  BOOL bReplaceAll=FALSE;
  int nResult;
  int i;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndFind=GetDlgItem(hDlg, IDC_SEARCH_FIND);
    hWndFindNextButton=GetDlgItem(hDlg, IDC_SEARCH_FIND_BUTTON);
    hWndMatchCase=GetDlgItem(hDlg, IDC_SEARCH_MATCHCASE);
    hWndWholeWord=GetDlgItem(hDlg, IDC_SEARCH_WHOLEWORD);
    hWndEscapeSeq=GetDlgItem(hDlg, IDC_SEARCH_ESCAPESEQ);
    hWndForward=GetDlgItem(hDlg, IDC_SEARCH_FORWARD);
    hWndBackward=GetDlgItem(hDlg, IDC_SEARCH_BACKWARD);
    hWndBeginning=GetDlgItem(hDlg, IDC_SEARCH_BEGINNING);
    if (bReplaceDlg == TRUE)
    {
      hWndReplace=GetDlgItem(hDlg, IDC_SEARCH_REPLACE);
      hWndReplaceButton=GetDlgItem(hDlg, IDC_SEARCH_REPLACE_BUTTON);
      hWndReplaceAllButton=GetDlgItem(hDlg, IDC_SEARCH_ALL_BUTTON);
      hWndInSelection=GetDlgItem(hDlg, IDC_SEARCH_INSEL);
    }

    FillComboboxSearch(hWndFind, hWndReplace);

    if (chrg.cpMin == chrg.cpMax || (ftflags & FR_SELECTION))
    {
      SendMessage(hWndFind, CB_SETCURSEL, 0, 0);
    }
    else
    {
      if (GetSelTextW(hWndEdit, &wszData))
      {
        API_SetWindowTextW(hWndFind, wszData);
        FreeTextExW(wszData);
      }
    }
    if (bReplaceDlg == TRUE) SendMessage(hWndReplace, CB_SETCURSEL, 0, 0);

    SendMessage(hWndForward, BM_SETCHECK, BST_UNCHECKED, 0);
    SendMessage(hWndBackward, BM_SETCHECK, BST_UNCHECKED, 0);
    SendMessage(hWndBeginning, BM_SETCHECK, BST_UNCHECKED, 0);
    if (bReplaceDlg == TRUE)
    {
      SendMessage(hWndInSelection, BM_SETCHECK, BST_UNCHECKED, 0);
      EnableWindow(hWndFindNextButton, TRUE);
      EnableWindow(hWndReplaceButton, TRUE);
    }
    if ((ftflags & FR_DOWN) && !(ftflags & FR_BEGINNING) && !(ftflags & FR_SELECTION))
      SendMessage(hWndForward, BM_SETCHECK, BST_CHECKED, 0);
    else if (ftflags & FR_UP)
      SendMessage(hWndBackward, BM_SETCHECK, BST_CHECKED, 0);
    else if (ftflags & FR_BEGINNING)
      SendMessage(hWndBeginning, BM_SETCHECK, BST_CHECKED, 0);
    else if (bReplaceDlg == TRUE && (ftflags & FR_SELECTION))
    {
      SendMessage(hWndInSelection, BM_SETCHECK, BST_CHECKED, 0);
      EnableWindow(hWndFindNextButton, FALSE);
      EnableWindow(hWndReplaceButton, FALSE);
    }
    SendMessage(hWndMatchCase, BM_SETCHECK, (BOOL)(ftflags & FR_MATCHCASE), 0);
    SendMessage(hWndWholeWord, BM_SETCHECK, (BOOL)(ftflags & FR_WHOLEWORD), 0);
    SendMessage(hWndEscapeSeq, BM_SETCHECK, (BOOL)(ftflags & FR_ESCAPESEQ), 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_SEARCH_REPLACE)
      return TRUE;
    else if (LOWORD(wParam) == IDC_SEARCH_MATCHCASE)
    {
      if (SendMessage(hWndMatchCase, BM_GETCHECK, 0, 0)) ftflags|=FR_MATCHCASE;
      else ftflags&=~FR_MATCHCASE;
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_WHOLEWORD)
    {
      if (SendMessage(hWndWholeWord, BM_GETCHECK, 0, 0)) ftflags|=FR_WHOLEWORD;
      else ftflags&=~FR_WHOLEWORD;
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_ESCAPESEQ)
    {
      if (SendMessage(hWndEscapeSeq, BM_GETCHECK, 0, 0)) ftflags|=FR_ESCAPESEQ;
      else ftflags&=~FR_ESCAPESEQ;
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_FORWARD)
    {
      ftflags&=~FR_UP&~FR_BEGINNING&~FR_SELECTION;
      ftflags|=FR_DOWN;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_BACKWARD)
    {
      ftflags&=~FR_DOWN&~FR_BEGINNING&~FR_SELECTION;
      ftflags|=FR_UP;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_BEGINNING)
    {
      ftflags&=~FR_UP&~FR_SELECTION;
      ftflags|=FR_DOWN|FR_BEGINNING;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_INSEL)
    {
      ftflags&=~FR_UP&~FR_BEGINNING;
      ftflags|=FR_DOWN|FR_SELECTION;
    }
    if (LOWORD(wParam) == IDC_SEARCH_FIND ||
        LOWORD(wParam) == IDC_SEARCH_FORWARD ||
        LOWORD(wParam) == IDC_SEARCH_BACKWARD ||
        LOWORD(wParam) == IDC_SEARCH_BEGINNING ||
        LOWORD(wParam) == IDC_SEARCH_INSEL)
    {
      EnableWindow(hWndFindNextButton, TRUE);
      if (bReplaceDlg == TRUE)
      {
        EnableWindow(hWndReplaceButton, TRUE);
        EnableWindow(hWndReplaceAllButton, TRUE);
      }

      if (HIWORD(wParam) == CBN_SELCHANGE)
      {
        if (bOldWindows)
          i=SendMessage(hWndFind, CB_GETLBTEXTLEN, (WPARAM)SendMessage(hWndFind, CB_GETCURSEL, 0, 0), 0);
        else
          i=SendMessageW(hWndFind, CB_GETLBTEXTLEN, (WPARAM)SendMessage(hWndFind, CB_GETCURSEL, 0, 0), 0);
      }
      else
      {
        if (bOldWindows)
          i=GetWindowTextLengthA(hWndFind);
        else
          i=GetWindowTextLengthW(hWndFind);
      }

      if (i == 0)
      {
        EnableWindow(hWndFindNextButton, FALSE);
        if (bReplaceDlg == TRUE)
        {
          EnableWindow(hWndReplaceButton, FALSE);
          EnableWindow(hWndReplaceAllButton, FALSE);
        }
      }
      else if (bReplaceDlg == TRUE && SendMessage(hWndInSelection, BM_GETCHECK, 0, 0))
      {
        EnableWindow(hWndFindNextButton, FALSE);
        EnableWindow(hWndReplaceButton, FALSE);
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      SaveComboboxSearch(hWndFind, hWndReplace);
      DestroyWindow(hDlg);
      hDlgModeless=0;
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_REPLACE_BUTTON)
    {
      bReplace=TRUE;
    }
    else if (LOWORD(wParam) == IDC_SEARCH_ALL_BUTTON)
    {
      bReplaceAll=TRUE;
    }
    if (LOWORD(wParam) == IDC_SEARCH_FIND_BUTTON ||
        LOWORD(wParam) == IDC_SEARCH_REPLACE_BUTTON ||
        LOWORD(wParam) == IDC_SEARCH_ALL_BUTTON)
    {
      FreeMemorySearch();

      if (!GetComboboxSearchText(hWndFind, &wszFind_orig, &wszFind))
      {
        FreeMemorySearch();
        return TRUE;
      }

      if (bReplaceDlg == TRUE)
      {
        if (!GetComboboxSearchText(hWndReplace, &wszReplace_orig, &wszReplace))
        {
          FreeMemorySearch();
          return TRUE;
        }
      }

      hFocus=GetFocus();
      SetFocus(hWndEdit);

      if (bReplaceDlg == TRUE && (bReplace == TRUE || bReplaceAll == TRUE))
        nResult=ReplaceTextW(hWndEdit, bReplaceAll);
      else
        nResult=FindTextW(hWndEdit);

      if (nResult == -1)
      {
        if (bReplaceAll == TRUE)
        {
          API_LoadString(hLangLib, MSG_REPLACE_COUNT, buf, BUFFER_SIZE);
          wsprintf(buf2, buf, nReplaceCount);
          MessageBox(hDlg, buf2, APP_SHORT_TITLE, MB_OK|MB_ICONINFORMATION);
        }
        else
        {
          API_LoadString(hLangLib, MSG_SEARCH_ENDED, buf, BUFFER_SIZE);
          MessageBox(hDlg, buf, APP_SHORT_TITLE, MB_OK|MB_ICONINFORMATION);
        }
      }
      else
      {
        if (bReplaceAll == FALSE && (ftflags & FR_BEGINNING))
        {
          ftflags&=~FR_BEGINNING;
          SendMessage(hWndBeginning, BM_SETCHECK, BST_UNCHECKED, 0);
          SendMessage(hWndForward, BM_SETCHECK, BST_CHECKED, 0);
        }
      }
      SetFocus(hFocus);
      return TRUE;
    }
  }
  return FALSE;
}

void FillComboboxSearch(HWND hWndFind, HWND hWndReplace)
{
  HWND hWnd;
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;
  int i;
  char szRegValue[32];
  unsigned char *szData;

  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Search", 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return;

  if (bOldWindows)
  {
    hWnd=hWndFind;
    lstrcpy(buf, "find%d");

    Read:
    for (i=0; i < nSearchStrings; ++i)
    {
      wsprintf(szRegValue, buf, i);

      if (RegQueryValueEx(hKey, szRegValue, NULL, &dwType, NULL, &dwSize) == ERROR_SUCCESS && dwSize > 0)
      {
        if (szData=(unsigned char *)API_HeapAlloc(hHeap, 0, dwSize + 1))
        {
          if (RegQueryValueEx(hKey, szRegValue, NULL, &dwType, (LPBYTE)szData, &dwSize) == ERROR_SUCCESS)
          {
            SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM)szData);
            API_HeapFree(hHeap, 0, (LPVOID)szData);
            continue;
          }
        }
      }
      break;
    }
    if (bReplaceDlg && hWnd != hWndReplace)
    {
      hWnd=hWndReplace;
      lstrcpy(buf, "replace%d");
      goto Read;
    }
  }
  else
  {
    hWnd=hWndFind;
    lstrcpyW((wchar_t *)buf, L"find%d");

    ReadW:
    for (i=0; i < nSearchStrings; ++i)
    {
      wsprintfW((wchar_t *)szRegValue, (wchar_t *)buf, i);

      if (RegQueryValueExW(hKey, (wchar_t *)szRegValue, NULL, &dwType, NULL, &dwSize) == ERROR_SUCCESS && dwSize > 0)
      {
        if (szData=(unsigned char *)API_HeapAlloc(hHeap, 0, dwSize + 1))
        {
          if (RegQueryValueExW(hKey, (wchar_t *)szRegValue, NULL, &dwType, (LPBYTE)szData, &dwSize) == ERROR_SUCCESS)
          {
            SendMessageW(hWnd, CB_ADDSTRING, 0, (LPARAM)szData);
            API_HeapFree(hHeap, 0, (LPVOID)szData);
            continue;
          }
        }
      }
      break;
    }
    if (bReplaceDlg && hWnd != hWndReplace)
    {
      hWnd=hWndReplace;
      lstrcpyW((wchar_t *)buf, L"replace%d");
      goto ReadW;
    }
  }
  RegCloseKey(hKey);
}

BOOL GetComboboxSearchText(HWND hWnd, wchar_t **wszText_orig, wchar_t **wszText)
{
  char *szData;
  int nIndex;
  int nTextLen;
  int nItemLen;

  if (bOldWindows)
    nTextLen=GetWindowTextLengthA(hWnd) + 1;
  else
    nTextLen=GetWindowTextLengthW(hWnd) + 1;

  if (*wszText_orig=*wszText=(wchar_t *)API_HeapAlloc(hHeap, 0, nTextLen * sizeof(wchar_t) + 2))
  {
    if (bOldWindows)
    {
      if (szData=(char *)API_HeapAlloc(hHeap, 0, nTextLen + 1))
      {
        GetWindowTextA(hWnd, szData, nTextLen);

        if (nSearchStrings)
        {
          if (*szData)
          {
            if ((nIndex=SendMessage(hWnd, CB_FINDSTRINGEXACT, 0, (LPARAM)szData)) != CB_ERR)
              SendMessage(hWnd, CB_DELETESTRING, (WPARAM)nIndex, 0);
          }
          else
          {
            for (nIndex=0; (nItemLen=SendMessage(hWnd, CB_GETLBTEXTLEN, nIndex, 0)) != CB_ERR && nItemLen; ++nIndex);
            if (!nItemLen) SendMessage(hWnd, CB_DELETESTRING, (WPARAM)nIndex, 0);
          }
          SendMessage(hWnd, CB_DELETESTRING, nSearchStrings - 1, 0);
          SendMessage(hWnd, CB_INSERTSTRING, 0, (LPARAM)szData);
          SendMessage(hWnd, CB_SETCURSEL, 0, 0);
        }
        MultiByteToWideChar(CP_ACP, 0, szData, nTextLen, *wszText_orig, nTextLen);
        API_HeapFree(hHeap, 0, (LPVOID)szData);
      }
      else return FALSE;
    }
    else
    {
      GetWindowTextW(hWnd, *wszText_orig, nTextLen);

      if (nSearchStrings)
      {
        if (**wszText_orig)
        {
          if ((nIndex=SendMessageW(hWnd, CB_FINDSTRINGEXACT, 0, (LPARAM)*wszText_orig)) != CB_ERR)
            SendMessage(hWnd, CB_DELETESTRING, (WPARAM)nIndex, 0);
        }
        else
        {
          for (nIndex=0; (nItemLen=SendMessageW(hWnd, CB_GETLBTEXTLEN, nIndex, 0)) != CB_ERR && nItemLen; ++nIndex);
          if (!nItemLen) SendMessage(hWnd, CB_DELETESTRING, (WPARAM)nIndex, 0);
        }
        SendMessage(hWnd, CB_DELETESTRING, nSearchStrings - 1, 0);
        SendMessageW(hWnd, CB_INSERTSTRING, 0, (LPARAM)*wszText_orig);
        SendMessage(hWnd, CB_SETCURSEL, 0, 0);
      }
    }

    if (ftflags & FR_ESCAPESEQ)
    {
      if (*wszText=(wchar_t *)API_HeapAlloc(hHeap, 0, nTextLen * sizeof(wchar_t) + 2))
      {
        if (!EscapeStringToEscapeDataW(*wszText_orig, *wszText))
        {
          API_LoadString(hLangLib, MSG_ERROR_SYNTAX, buf, BUFFER_SIZE);
          MessageBox(GetParent(hWnd), buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
          return FALSE;
        }
      }
      else return FALSE;
    }
  }
  else return FALSE;

  return TRUE;
}

void SaveComboboxSearch(HWND hWndFind, HWND hWndReplace)
{
  HWND hWnd;
  HKEY hKey;
  int nSize;
  int i;
  char szRegValue[32];
  unsigned char *szData;

  if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Search", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) != ERROR_SUCCESS)
    return;

  if (bOldWindows)
  {
    hWnd=hWndFind;
    lstrcpy(buf, "find%d");

    Save:
    for (i=0; i < nSearchStrings; ++i)
    {
      wsprintf(szRegValue, buf, i);

      if ((nSize=SendMessage(hWnd, CB_GETLBTEXTLEN, (WPARAM)i, 0)) != CB_ERR)
      {
        ++nSize;

        if (szData=(unsigned char *)API_HeapAlloc(hHeap, 0, nSize + 1))
        {
          SendMessage(hWnd, CB_GETLBTEXT, (WPARAM)i, (LPARAM)szData);
          RegSetValueEx(hKey, szRegValue, 0, REG_SZ, (LPBYTE)szData, nSize);
          API_HeapFree(hHeap, 0, (LPVOID)szData);
          continue;
        }
      }
      break;
    }
    if (bReplaceDlg && hWnd != hWndReplace)
    {
      hWnd=hWndReplace;
      lstrcpy(buf, "replace%d");
      goto Save;
    }
  }
  else
  {
    hWnd=hWndFind;
    lstrcpyW((wchar_t *)buf, L"find%d");

    SaveW:
    for (i=0; i < nSearchStrings; ++i)
    {
      wsprintfW((wchar_t *)szRegValue, (wchar_t *)buf, i);

      if ((nSize=SendMessageW(hWnd, CB_GETLBTEXTLEN, (WPARAM)i, 0)) != CB_ERR)
      {
        ++nSize;

        if (szData=(unsigned char *)API_HeapAlloc(hHeap, 0, nSize * sizeof(wchar_t) + 2))
        {
          SendMessageW(hWnd, CB_GETLBTEXT, (WPARAM)i, (LPARAM)szData);
          RegSetValueExW(hKey, (wchar_t *)szRegValue, 0, REG_SZ, (LPBYTE)szData, nSize * sizeof(wchar_t));
          API_HeapFree(hHeap, 0, (LPVOID)szData);
          continue;
        }
      }
      break;
    }
    if (bReplaceDlg && hWnd != hWndReplace)
    {
      hWnd=hWndReplace;
      lstrcpyW((wchar_t *)buf, L"replace%d");
      goto SaveW;
    }
  }
  RegCloseKey(hKey);
}

int FindTextW(HWND hWnd)
{
  FINDTEXTEXW ftW;
  int nResult;

  if ((ftflags & FR_DOWN) && !(ftflags&FR_BEGINNING) && !(ftflags&FR_SELECTION))
  {
    ftW.chrg.cpMin=chrg.cpMax;
    ftW.chrg.cpMax=-1;
  }
  else if (ftflags & FR_UP)
  {
    ftW.chrg.cpMin=chrg.cpMin;
    ftW.chrg.cpMax=0;
  }
  else if (ftflags & FR_BEGINNING)
  {
    ftW.chrg.cpMin=0;
    ftW.chrg.cpMax=-1;
  }
  ftW.lpstrText=wszFind;

  if ((nResult=SendMessage(hWnd, EM_FINDTEXTEXW, ftflags, (LPARAM)&ftW)) != -1)
  {
    //EM_FINDTEXTEX bug workarround
    if (ftW.chrgText.cpMax - ftW.chrgText.cpMin == 1)
    {
      TEXTRANGE txtrng;
      txtrng.lpstrText=buf;
      txtrng.chrg.cpMin=ftW.chrgText.cpMin;
      txtrng.chrg.cpMax=ftW.chrgText.cpMax;
      SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
      if (*buf == 0x00) nResult=-1;
    }
    if (nResult != -1) SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&ftW.chrgText);
  }
  return nResult;
}

int ReplaceTextW(HWND hWnd, BOOL all)
{
  FINDTEXTEXW ftW;
  CHARRANGE crInitial;
  CHARRANGE crSelection;
  CHARRANGE crPrevious;
  wchar_t *wszData;
  BOOL bFirstCall=TRUE;
  int nFirstLine;
  int nCurrentLine;
  int nResult;
  int i;

  nReplaceCount=0;

  if (all)
  {
    crInitial.cpMin=chrg.cpMin;
    crInitial.cpMax=chrg.cpMax;
    nFirstLine=SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0);
  }
  if (ftflags & FR_SELECTION)
  {
    crSelection.cpMin=chrg.cpMin;
    crSelection.cpMax=chrg.cpMax;
  }
  else if (!(all && (ftflags & FR_BEGINNING)))
  {
    if (chrg.cpMin != chrg.cpMax)
    {
      if (GetSelTextW(hWnd, &wszData))
      {
        if (((ftflags & FR_MATCHCASE) && !API_lstrcmpW(wszFind, wszData)) ||
            (!(ftflags & FR_MATCHCASE) && !API_lstrcmpiW(wszFind, wszData)))
        {
          FreeTextExW(wszData);
          goto Replace;
        }
        FreeTextExW(wszData);
      }
      else return -1;
    }
  }
  goto Find;

  do
  {
    Replace:
    crPrevious.cpMin=chrg.cpMin;
    crPrevious.cpMax=chrg.cpMax;

    ++nReplaceCount;
    ReplaceSelW(hWnd, wszReplace);
    i=API_lstrlenW(wszFind) - API_lstrlenW(wszReplace);

    if (all)
    {
      if (crInitial.cpMin >= crPrevious.cpMax) crInitial.cpMin-=i;
      else if (crInitial.cpMin < crPrevious.cpMax && crInitial.cpMin > crPrevious.cpMin) crInitial.cpMin=crPrevious.cpMin;

      if (crInitial.cpMax >= crPrevious.cpMax) crInitial.cpMax-=i;
      else if (crInitial.cpMax < crPrevious.cpMax && crInitial.cpMax > crPrevious.cpMin) crInitial.cpMax=nPosition;
    }
    if (ftflags & FR_SELECTION) crSelection.cpMax-=i;
    else if (ftflags & FR_UP) SendMessage(hWnd, EM_SETSEL, crPrevious.cpMin, crPrevious.cpMin);

    Find:
    if (ftflags & FR_SELECTION)
    {
      ftW.chrg.cpMin=(bFirstCall)?crSelection.cpMin:chrg.cpMax;
      ftW.chrg.cpMax=crSelection.cpMax;
    }
    else if (ftflags & FR_BEGINNING)
    {
      ftW.chrg.cpMin=(bFirstCall)?0:chrg.cpMax;
      ftW.chrg.cpMax=-1;
    }
    else if (ftflags & FR_DOWN)
    {
      ftW.chrg.cpMin=chrg.cpMax;
      ftW.chrg.cpMax=-1;
    }
    else if (ftflags & FR_UP)
    {
      ftW.chrg.cpMin=chrg.cpMin;
      ftW.chrg.cpMax=0;
    }
    bFirstCall=FALSE;
    ftW.lpstrText=wszFind;

    if ((nResult=SendMessage(hWnd, EM_FINDTEXTEXW, ftflags, (LPARAM)&ftW)) != -1)
    {
      //EM_FINDTEXTEX bug workarround
      if (ftW.chrgText.cpMax - ftW.chrgText.cpMin == 1)
      {
        TEXTRANGE txtrng;
        txtrng.lpstrText=buf;
        txtrng.chrg.cpMin=ftW.chrgText.cpMin;
        txtrng.chrg.cpMax=ftW.chrgText.cpMax;
        SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
        if (*buf == 0x00) nResult=-1;
      }
      if (nResult != -1) SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&ftW.chrgText);
    }
  }
  while (all && nResult != -1);

  if (all)
  {
    SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&crInitial);
    nCurrentLine=SendMessage(hWnd, EM_GETFIRSTVISIBLELINE, 0, 0);
    if (bOldRichEdit)
    {
      SendMessage(hWnd, EM_LINESCROLL, 0, -1);
      SendMessage(hWnd, EM_LINESCROLL, 0, nFirstLine);
    }
    else SendMessage(hWnd, EM_LINESCROLL, 0, nFirstLine - nCurrentLine);
  }
  return nResult;
}

BOOL EscapeStringToEscapeDataW(wchar_t *input, wchar_t *output)
{
  wchar_t *a=input;
  wchar_t *b=output;
  unsigned char *c;
  char hex[2];

  for (hex[2]='\0'; *a; ++b, ++a)
  {
    if (*a == '\\')
    {
      if (*++a == '\\') *b='\\';
      else if (*a == 'n') *b='\r';
      else if (*a == 't') *b='\t';
      else if (*a == '[')
      {
        --b, ++a;

        do
        {
          c=(unsigned char *)++b;
          hex[0]=(char)*a;
          if (*a == 0) goto Error;
          hex[1]=(char)*++a;
          if (*a == 0) goto Error;
          *c=hex2dec(hex);
          if (*c == -1) goto Error;
          while (*++a == ' ');

          ++c;
          hex[0]=(char)*a;
          if (*a == 0) goto Error;
          hex[1]=(char)*++a;
          if (*a == 0) goto Error;
          *c=hex2dec(hex);
          if (*c == -1) goto Error;
          while (*++a == ' ');
        }
        while (*a && *a != ']');
      }
      else goto Error;
    }
    else *b=*a;
  }
  *b='\0';
  return TRUE;

  Error:
  *output='\0';
  return FALSE;
}

void EscapeDataToEscapeStringW(wchar_t *input, wchar_t *output)
{
  wchar_t *a;
  wchar_t *b;

  for (a=input, b=output; *a; ++b, ++a)
  {
    if (*a == '\t') *b='\\', *++b='t';
    else if (*a == '\r') *b='\\', *++b='n';
    else if (*a == '\\') *b='\\', *++b='\\';
    else *b=*a;
  }
  *b='\0';
}


//// Paste operation

void ReplaceSelA(HWND hWnd, char *pData)
{
  wchar_t *pDataW;
  int nLen;
  int nLenW;

  nLen=lstrlen(pData) + 1;
  nLenW=nLen * sizeof(wchar_t);

  if (pDataW=(wchar_t *)API_HeapAlloc(hHeap, 0, nLenW + 2))
  {
    MultiByteToWideChar(CP_ACP, 0, pData, nLen, pDataW, nLenW);
    ReplaceSelW(hWnd, pDataW);
    API_HeapFree(hHeap, 0, (LPVOID)pDataW);
  }
}

void ReplaceSelW(HWND hWnd, wchar_t *pData)
{
  EDITSTREAM es;
  PASTEDATA pd;

  pd.pData=(unsigned char *)pData;
  pd.nLength=0;
  es.dwCookie=(DWORD)&pd;
  es.pfnCallback=ReplaceSelCallbackW;
  SendMessage(hWnd, EM_STREAMIN, SF_TEXT|SF_UNICODE|SFF_SELECTION, (LPARAM)&es);
}

DWORD CALLBACK ReplaceSelCallbackW(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
  PASTEDATA *lpData=(PASTEDATA *)dwCookie;
  wchar_t *pSrc=(wchar_t *)lpData->pData;
  wchar_t *pDest=(wchar_t *)pbBuff;
  int i;

  for (i=0, cb=cb / sizeof(wchar_t) - 1; i < cb && pSrc[lpData->nLength]; ++i)
  {
    pDest[i]=pSrc[lpData->nLength++];

    if (pDest[i] == '\r' && pSrc[lpData->nLength] != '\n')
      pDest[++i]='\n';
  }

  *pcb=i*sizeof(wchar_t);

  return 0;
}

int GetSelTextW(HWND hWnd, wchar_t **pText)
{
  CHARRANGE cr;
  int nLen;
  int nLenW;

  if (chrg.cpMin == chrg.cpMax) goto Error;
  cr.cpMin=chrg.cpMin;
  cr.cpMax=chrg.cpMax;
  nLen=(cr.cpMax - cr.cpMin) + 1;
  nLenW=nLen * sizeof(wchar_t);

  if (bOldRichEdit)
  {
    TEXTRANGE txtrng;
    int nLayout;

    if (*pText=(wchar_t *)API_HeapAlloc(hHeap, 0, nLenW + 2))
    {
      if (txtrng.lpstrText=(char *)API_HeapAlloc(hHeap, 0, nLen + 1))
      {
        txtrng.chrg.cpMin=cr.cpMin;
        txtrng.chrg.cpMax=cr.cpMax;

        nLayout=SetLayout();
        SendMessage(hWnd, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
        RestoreLayout(nLayout);

        MultiByteToWideChar(CP_ACP, 0, txtrng.lpstrText, nLen, *pText, nLen);
        API_HeapFree(hHeap, 0, (LPVOID)txtrng.lpstrText);
        return nLen;
      }
      else API_HeapFree(hHeap, 0, (LPVOID)*pText);
    }
  }
  else
  {
    GETTEXTEX gt;

    if (*pText=(wchar_t *)API_HeapAlloc(hHeap, 0, nLenW + 2))
    {
      gt.cb=nLenW;
      gt.flags=GT_SELECTION;
      gt.codepage=1200;
      gt.lpDefaultChar=NULL;
      gt.lpUsedDefChar=NULL;
      SendMessage(hWnd, EM_GETTEXTEX, (WPARAM)&gt, (LPARAM)*pText);

      return nLen;
    }
  }

  Error:
  *pText=NULL;
  return 0;
}

void FreeTextExW(wchar_t *pText)
{
  if (pText) API_HeapFree(hHeap, 0, (LPVOID)pText);
}


//// Go to line

BOOL CALLBACK GoToLineDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  int nLine;
  int nColumn;
  int nLineNumber;
  int nNewLine;
  int nNewLineLen;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    nLine=SendMessage(hWndEdit, EM_EXLINEFROMCHAR, 0, nPosition)+1;
    SetDlgItemInt(hDlg, IDC_GOTOLINE, nLine, FALSE);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      nLineNumber=GetDlgItemInt(hDlg, IDC_GOTOLINE, NULL, TRUE);

      if (!nLineNumber)
      {
        API_LoadString(hLangLib, MSG_WRONG_STRING, buf, BUFFER_SIZE);
        MessageBox(hDlg, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
        return FALSE;
      }
      else if (nLineNumber < 0)
      {
        nLineNumber=SendMessage(hWndEdit, EM_GETLINECOUNT, 0, 0) + nLineNumber + 1;
        if (nLineNumber <= 0) nLineNumber=1;
      }

      nLine=SendMessage(hWndEdit, EM_EXLINEFROMCHAR, 0, nPosition) + 1;
      nColumn=SendMessage(hWndEdit, EM_LINEINDEX, nLine - 1, 0);
      nColumn=nPosition - nColumn + 1;
      nNewLine=SendMessage(hWndEdit, EM_LINEINDEX, nLineNumber - 1, 0);
      if (nNewLine == -1) nNewLine=SendMessage(hWndEdit, EM_LINEINDEX, SendMessage(hWndEdit, EM_GETLINECOUNT, 0, 0) - 1, 0);
      nNewLineLen=SendMessage(hWndEdit, EM_LINELENGTH, nNewLine, 0);

      if (nColumn <= nNewLineLen) nNewLine+=nColumn - 1;
      else nNewLine+=nNewLineLen;
      SendMessage(hWndEdit, EM_SETSEL, nNewLine, nNewLine);
    }
    if (LOWORD(wParam) == IDOK ||
        LOWORD(wParam) == IDCANCEL)
    {
      DestroyWindow(hDlg);
      hDlgModeless=0;
      return TRUE;
    }
  }
  return FALSE;
}


//// Recent files

void RecentFilesZero()
{
  if (szRecentNames) memset(szRecentNames, 0, nRecentFiles * MAX_PATH);
  if (szRecentPositions) memset(szRecentPositions, 0, nRecentFiles * sizeof(DWORD));
  if (szRecentCodepages) memset(szRecentCodepages, 0, nRecentFiles * sizeof(DWORD));
}

void RecentFilesRead()
{
  char szRegValue[32];
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;
  int i;

  //Read recent files array
  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Recent", 0, KEY_READ, &hKey) != ERROR_SUCCESS)
    return;

  for (i=0; i < nRecentFiles; ++i)
  {
    wsprintf(szRegValue, "name%d", i);
    dwSize=MAX_PATH;
    if (RegQueryValueEx(hKey, szRegValue, NULL, &dwType, (LPBYTE)szRecentNames[i], &dwSize) != ERROR_SUCCESS)
      break;

    wsprintf(szRegValue, "pos%d", i);
    dwSize=sizeof(DWORD);
    RegQueryValueEx(hKey, szRegValue, NULL, &dwType, (LPBYTE)&szRecentPositions[i], &dwSize);

    wsprintf(szRegValue, "cp%d", i);
    dwSize=sizeof(DWORD);
    RegQueryValueEx(hKey, szRegValue, NULL, &dwType, (LPBYTE)&szRecentCodepages[i], &dwSize);
  }
  RegCloseKey(hKey);
}

BOOL RecentFilesGet(char *file, int *position, int *codepage)
{
  int i;

  if (!*file) return FALSE;

  for (i=0; i < nRecentFiles; ++i)
    if (!lstrcmpi(szRecentNames[i], file))
    {
      if (position) *position=szRecentPositions[i];
      if (codepage) *codepage=szRecentCodepages[i];
      return TRUE;
    }

  return FALSE;
}

BOOL RecentFilesUpdate(char *file, int position, int codepage)
{
  int i;

  if (!*file) return FALSE;

  //Update recent files array - move/add current file to the first place
  for (i=0; i < nRecentFiles; ++i)
    if (!lstrcmpi(szRecentNames[i], file))
    {
      if (position == -1) position=szRecentPositions[i];
      if (codepage == -1) codepage=szRecentCodepages[i];
      ++i;
      break;
    }

  if (position == -1) position=0;
  if (codepage == -1) codepage=nDefaultCodePage;

  if (--i)
    for (; i >= 0; --i)
    {
      lstrcpy(szRecentNames[i], szRecentNames[i-1]);
      szRecentPositions[i]=szRecentPositions[i-1];
      szRecentCodepages[i]=szRecentCodepages[i-1];
    }

  lstrcpy(szRecentNames[0], file);
  szRecentPositions[0]=position;
  szRecentCodepages[0]=codepage;

  return TRUE;
}

int RecentFilesDeleteDead()
{
  int i;
  int a;
  int nDead=0;

  //Delete files from recent files array if they doesn't exist
  for (i=0; i < nRecentFiles && *szRecentNames[i]; ++i)
  {
    while (*szRecentNames[i] && !FileExists(szRecentNames[i]))
    {
      for (a=i; a < nRecentFiles-1; ++a)
      {
        lstrcpy(szRecentNames[a], szRecentNames[a+1]);
        szRecentPositions[a]=szRecentPositions[a+1];
        szRecentCodepages[a]=szRecentCodepages[a+1];
      }
      *szRecentNames[nRecentFiles-1]='\0';
      szRecentPositions[nRecentFiles-1]=0;
      szRecentCodepages[nRecentFiles-1]=0;
      ++nDead;
    }
  }
  return nDead;
}

void RecentFilesSave()
{
  char szRegValue[32];
  HKEY hKey;
  int i;

  //Save recent files array
  if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Recent", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) != ERROR_SUCCESS)
    return;

  for (i=0; i < nRecentFiles; ++i)
  {
    wsprintf(szRegValue, "name%d", i);
    RegSetValueEx(hKey, szRegValue, 0, REG_SZ, (LPBYTE)(*szRecentNames[i]?szRecentNames[i]:""), *szRecentNames[i]?lstrlen(szRecentNames[i])+1:1);

    wsprintf(szRegValue, "pos%d", i);
    RegSetValueEx(hKey, szRegValue, 0, REG_DWORD, (LPBYTE)&szRecentPositions[i], sizeof(DWORD));

    wsprintf(szRegValue, "cp%d", i);
    RegSetValueEx(hKey, szRegValue, 0, REG_DWORD, (LPBYTE)&szRecentCodepages[i], sizeof(DWORD));
  }
  RegCloseKey(hKey);
}

void RecentFilesMenu()
{
  int i;

  for (i=1; DeleteMenu(hMainMenu, IDM_RECENT_FILES+i, MF_BYCOMMAND); ++i);

  if (nRecentFiles && *szRecentNames[0])
  {
    API_LoadString(hLangLib, STR_RECENT_FILES_DELETE_DEAD, buf, BUFFER_SIZE);
    ModifyMenu(hMainMenu, IDM_RECENT_FILES, MF_BYCOMMAND|MF_STRING, IDM_RECENT_FILES, buf);

    for (i=0; *szRecentNames[i] && i < nRecentFiles; ++i)
    {
      TrimPath(buf, szRecentNames[i], RECENTFILES_RECORD_LENGTH);
      InsertMenu(hMainMenu, IDM_RECENT_FILES, MF_BYCOMMAND|MF_STRING, IDM_RECENT_FILES+i+1, buf);
    }
    InsertMenu(hMainMenu, IDM_RECENT_FILES, MF_BYCOMMAND|MF_SEPARATOR, IDM_RECENT_FILES+i+1, NULL);
  }
  else
  {
    API_LoadString(hLangLib, STR_RECENT_FILES_EMPTY, buf, BUFFER_SIZE);
    ModifyMenu(hMainMenu, IDM_RECENT_FILES, MF_BYCOMMAND|MF_STRING|MF_GRAYED, IDM_RECENT_FILES, buf);
  }
}

int TrimPath(char *szResult, char *pPath, int nMax)
{
  int nFileLen;
  int nLen=lstrlen(pPath);
  char *pEnd=pPath + nLen;
  char *pFile=pEnd - 1;

  if (nLen <= nMax)
  {
    lstrcpy(szResult, pPath);
    return nLen;
  }

  while (pFile >= pPath && *pFile != '\\') --pFile;

  if (pFile >= pPath)
  {
    nFileLen=pEnd - pFile;

    if (nMax > nFileLen + 2)
    {
      lstrcpyn(szResult, pPath, nMax - nFileLen - 2);
      lstrcat(szResult, "...");
      lstrcat(szResult, pFile);
      return nMax;
    }
  }
  lstrcpy(szResult, "...");
  lstrcat(szResult, pEnd - nMax + 3);
  return nMax;
}


//// Language menu

void LanguageMenu()
{
  WIN32_FIND_DATA wfd;
  HANDLE hFind;
  UINT uFlags;
  int i;
  nLangModules=0;

  for (i=1; DeleteMenu(hMainMenu, IDM_LANGUAGE+i, MF_BYCOMMAND); ++i);

  wsprintf(buf, "%s\\AkelFiles\\Langs\\*.dll", szExeDir);

  if ((hFind=FindFirstFile(buf, &wfd)) != INVALID_HANDLE_VALUE)
  {
    i=1;

    do
    {
      if (!lstrcmpi(szLangModule, wfd.cFileName))
        uFlags=MF_BYCOMMAND|MF_STRING|MF_CHECKED;
      else
        uFlags=MF_BYCOMMAND|MF_STRING|MF_UNCHECKED;
      GetBaseName(wfd.cFileName, buf);
      InsertMenu(hMainMenu, IDM_LANGUAGE, uFlags, IDM_LANGUAGE+i, buf);
      ++i;
    }
    while (FindNextFile(hFind, &wfd));

    nLangModules=i-1;
    InsertMenu(hMainMenu, IDM_LANGUAGE, MF_BYCOMMAND|MF_SEPARATOR, IDM_LANGUAGE+i, NULL);
    FindClose(hFind);
  }

  API_LoadString(hLangLib, STR_LANGUAGE_NONE, buf, BUFFER_SIZE);
  if (!*szLangModule)
    uFlags=MF_BYCOMMAND|MF_STRING|MF_CHECKED;
  else
    uFlags=MF_BYCOMMAND|MF_STRING|MF_UNCHECKED;
  ModifyMenu(hMainMenu, IDM_LANGUAGE, uFlags, IDM_LANGUAGE, buf);
}


//// Word breaking

BOOL isInDelimiterList(wchar_t c)
{
  if (wcschr(wszDelimiters, c) != NULL)
    return TRUE;
  else
    return FALSE;
}

int CALLBACK EditWordBreakProc(LPTSTR pchText, int ichCurrent, int cch, int code)
{
  int lRet=0;
  wchar_t *pchTextW=(wchar_t *)pchText;
  int cchW=cch/sizeof(wchar_t);

  if (code == WB_LEFTBREAK || code == WB_MOVEWORDLEFT)
  {
    // Finds the next character that begins a word before the specified position.
    {
      if (bFirstWordBreak)
      {
        do
          --ichCurrent;
        while (isInDelimiterList(pchTextW[ichCurrent]));
      }

      for (--ichCurrent; ichCurrent >= 0 && !isInDelimiterList(pchTextW[ichCurrent]); --ichCurrent);

      if (ichCurrent >= 0)
      {
        bFirstWordBreak=TRUE;
        lRet=ichCurrent+1;
      }
      else
      {
        bFirstWordBreak=FALSE;
        lRet=-1;
      }
    }
  }
  else if (code == WB_MOVEWORDRIGHT)
  {
    //Finds the next character that ends a word after the specified position.
    {
      if (bFirstWordBreak)
      {
        while (isInDelimiterList(pchTextW[ichCurrent]))
          ++ichCurrent;
        ++ichCurrent;
      }

      while (ichCurrent < cchW && !isInDelimiterList(pchTextW[ichCurrent]))
        ++ichCurrent;

      if (ichCurrent < cchW)
      {
        bFirstWordBreak=TRUE;
        lRet=ichCurrent;
      }
      else
      {
        bFirstWordBreak=FALSE;
        lRet=-1;
      }
    }
  }
  else if (code == WB_CLASSIFY)
  {
    //Retrieves the character class and word break flags of the character at the specified position.
    {
      if (!isInDelimiterList(pchTextW[ichCurrent]))
        lRet=WBF_ISWHITE; //WBF_BREAKAFTER|WBF_BREAKLINE;
      else
        lRet=-1;
    }
  }
  return lRet;
}


//// Recode

void RecodeText(HWND hWnd, int nCodePage)
{
  CHARRANGE cr;
  wchar_t *wszText;
  char *szText;
  BOOL bAllFile=FALSE;
  int i;

  cr.cpMin=chrg.cpMin;
  cr.cpMax=chrg.cpMax;

  if (cr.cpMin == cr.cpMax)
  {
    SendMessage(hWnd, EM_SETSEL, 0, -1);
    bAllFile=TRUE;
  }

  if (i=GetSelTextW(hWnd, &wszText))
  {
    if (szText=(char *)API_HeapAlloc(hHeap, 0, i + 1))
    {
      WideCharToMultiByte(CP_ACP, 0, wszText, i, szText, i, NULL, NULL);

      if (nCodePage == CP_UNICODE_UTF8)
        i=UTF8toUTF16(szText, (unsigned int *)&i, wszText, i);
      else
        MultiByteToWideChar(nCodePage, 0, szText, i, wszText, i);

      if (bAllFile == FALSE)
        cr.cpMax=cr.cpMin + i - 1;

      ReplaceSelW(hWnd, wszText);
      API_HeapFree(hHeap, 0, (LPVOID)szText);
    }
    FreeTextExW(wszText);
  }
  SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
}

BOOL CALLBACK RecodeDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndCP;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndCP=GetDlgItem(hDlg, IDC_RECODE_CODEPAGE);

    FillComboboxCodepage(hWndCP);
    SelectComboboxCodepage(hWndCP, nRecodeCodePage);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK)
    {
      nRecodeCodePage=GetComboboxCodepage(hWndCP);
      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      bRecodeDlg=FALSE;
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  return FALSE;
}


//// Options

LRESULT CALLBACK CBTProc(int iCode, WPARAM wParam, LPARAM lParam)
{
  //Center PropertySheet
  if (iCode == HCBT_ACTIVATE)
  {
    RECT rcEdit;
    RECT rcSheet;

    UnhookWindowsHookEx(cbtHook);

    GetWindowRect(hMainWnd, &rcEdit);
    GetWindowRect((HWND)wParam, &rcSheet);
    rcSheet.left=rcEdit.left + ((rcEdit.right - rcEdit.left) / 2) - ((rcSheet.right - rcSheet.left) / 2);
    rcSheet.top=rcEdit.top + ((rcEdit.bottom - rcEdit.top) / 2) - ((rcSheet.bottom - rcSheet.top) / 2);
    if (rcSheet.left < 0) rcSheet.left=0;
    if (rcSheet.top < 0) rcSheet.top=0;

    SetWindowPos((HWND)wParam, NULL, rcSheet.left, rcSheet.top, 0, 0, SWP_NOSIZE|SWP_NOZORDER);
  }
  return 0;
}

int CALLBACK PropSheetProc(HWND hDlg, UINT uMsg, LPARAM lParam)
{
  //Remove "?"
  if (uMsg == PSCB_PRECREATE)
    ((DLGTEMPLATE *)lParam)->style&=~DS_CONTEXTHELP;

  return TRUE;
}

BOOL CALLBACK OptionsGeneralDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndCP;
  HWND hWndRecentFiles;
  HWND hWndRecentFilesSpin;
  HWND hWndRecentFilesClear;
  HWND hWndSearchStrings;
  HWND hWndSearchStringsSpin;
  HWND hWndSearchStringsClear;
  HWND hWndExecute;
  HKEY hKey;
  int i;

  if (uMsg == WM_INITDIALOG)
  {
    hWndRecentFiles=GetDlgItem(hDlg, IDC_OPTIONS_RECENTFILES_AMOUNT);
    hWndRecentFilesSpin=GetDlgItem(hDlg, IDC_OPTIONS_RECENTFILES_AMOUNT_SPIN);
    hWndRecentFilesClear=GetDlgItem(hDlg, IDC_OPTIONS_RECENTFILES_CLEAR);
    hWndSearchStrings=GetDlgItem(hDlg, IDC_OPTIONS_SEARCHSTRINGS_AMOUNT);
    hWndSearchStringsSpin=GetDlgItem(hDlg, IDC_OPTIONS_SEARCHSTRINGS_AMOUNT_SPIN);
    hWndSearchStringsClear=GetDlgItem(hDlg, IDC_OPTIONS_SEARCHSTRINGS_CLEAR);
    hWndExecute=GetDlgItem(hDlg, IDC_OPTIONS_EXEC);
    hWndCP=GetDlgItem(hDlg, IDC_OPTIONS_CODEPAGE);

    SendMessage(hWndExecute, EM_LIMITTEXT, (WPARAM)BUFFER_SIZE, 0);
    SendMessage(hWndRecentFilesSpin, UDM_SETBUDDY, (WPARAM)hWndRecentFiles, 0);
    SendMessage(hWndRecentFilesSpin, UDM_SETRANGE, 0, (LPARAM)MAKELONG(999, 0));
    SendMessage(hWndSearchStringsSpin, UDM_SETBUDDY, (WPARAM)hWndSearchStrings, 0);
    SendMessage(hWndSearchStringsSpin, UDM_SETRANGE, 0, (LPARAM)MAKELONG(999, 0));

    SetDlgItemInt(hDlg, IDC_OPTIONS_RECENTFILES_AMOUNT, nRecentFiles, FALSE);
    if (!nRecentFiles || !*szRecentNames[0]) EnableWindow(hWndRecentFilesClear, FALSE);
    SetDlgItemInt(hDlg, IDC_OPTIONS_SEARCHSTRINGS_AMOUNT, nSearchStrings, FALSE);
    if (!nSearchStrings) EnableWindow(hWndSearchStringsClear, FALSE);
    SetWindowText(hWndExecute, szCommand);

    FillComboboxCodepage(hWndCP);
    SelectComboboxCodepage(hWndCP, nDefaultCodePage);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_OPTIONS_RECENTFILES_CLEAR)
    {
      RecentFilesZero();
      RecentFilesSave();
      RecentFilesMenu();
      EnableWindow((HWND)lParam, FALSE);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OPTIONS_SEARCHSTRINGS_CLEAR)
    {
      if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad", 0, KEY_ENUMERATE_SUB_KEYS, &hKey) == ERROR_SUCCESS)
      {
        RegDeleteKey(hKey, "Search");
        RegCloseKey(hKey);

        if (RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad\\Search", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) == ERROR_SUCCESS)
          RegCloseKey(hKey);
      }
      EnableWindow((HWND)lParam, FALSE);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OPTIONS_EXEC_BROWSE)
    {
      OPENFILENAME efn={0};

      buf[0]='\0';
      efn.lStructSize      =sizeof(OPENFILENAME);
      efn.hwndOwner        =hDlg;
      efn.lpstrFile        =(char *)&buf;
      efn.nMaxFile         =MAX_PATH;
      efn.lpstrFilter      ="Executable files (*.exe)\0*.exe\0All files (*.*)\0*.*";
      efn.nFilterIndex     =1;
      efn.Flags            =OFN_HIDEREADONLY|OFN_PATHMUSTEXIST;

      if (GetOpenFileName(&efn))
      {
        wsprintf(buf2, "\"%s\"", buf);
        SetDlgItemText(hDlg, IDC_OPTIONS_EXEC, buf2);
      }
      return TRUE;
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (((NMHDR *)lParam)->code == PSN_APPLY)
    {
      //Recent files
      i=GetDlgItemInt(hDlg, IDC_OPTIONS_RECENTFILES_AMOUNT, NULL, FALSE);
      if (nRecentFiles != i)
      {
        nRecentFiles=i;
        FreeMemoryRecentFiles();

        if (nRecentFiles)
        {
          if ((szRecentNames=(char (*)[MAX_PATH])API_HeapAlloc(hHeap, 0, nRecentFiles * MAX_PATH)) &&
              (szRecentPositions=(DWORD *)API_HeapAlloc(hHeap, 0, nRecentFiles * sizeof(DWORD))) &&
              (szRecentCodepages=(DWORD *)API_HeapAlloc(hHeap, 0, nRecentFiles * sizeof(DWORD))))
          {
            RecentFilesZero();
            RecentFilesRead();
          }
        }
        RecentFilesMenu();

        if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Akelsoft\\AkelPad", 0, KEY_ENUMERATE_SUB_KEYS, &hKey) == ERROR_SUCCESS)
        {
          RegDeleteKey(hKey, "Recent");
          RegCloseKey(hKey);
          RecentFilesSave();
        }
      }

      //Search history
      nSearchStrings=GetDlgItemInt(hDlg, IDC_OPTIONS_SEARCHSTRINGS_AMOUNT, NULL, FALSE);

      //Execute command
      GetDlgItemText(hDlg, IDC_OPTIONS_EXEC, szCommand, BUFFER_SIZE);

      //Default codepage
      nDefaultCodePage=GetComboboxCodepage(hWndCP);
    }
  }
  return FALSE;
}

BOOL CALLBACK OptionsAdvancedDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndFileTypesOpen;
  static HWND hWndAssociateOpen;
  static HWND hWndFileTypesEdit;
  static HWND hWndAssociateEdit;
  static HWND hWndFileTypesPrint;
  static HWND hWndAssociatePrint;
  HWND hWndMarginLeft;
  HWND hWndMarginLeftSpin;
  HWND hWndMarginRight;
  HWND hWndMarginRightSpin;
  HWND hWndDelimiters;
  BOOL bState;

  if (uMsg == WM_INITDIALOG)
  {
    hWndFileTypesOpen=GetDlgItem(hDlg, IDC_OPTIONS_FILETYPES_OPEN);
    hWndAssociateOpen=GetDlgItem(hDlg, IDC_OPTIONS_ASSOCIATE_OPEN);
    hWndFileTypesEdit=GetDlgItem(hDlg, IDC_OPTIONS_FILETYPES_EDIT);
    hWndAssociateEdit=GetDlgItem(hDlg, IDC_OPTIONS_ASSOCIATE_EDIT);
    hWndFileTypesPrint=GetDlgItem(hDlg, IDC_OPTIONS_FILETYPES_PRINT);
    hWndAssociatePrint=GetDlgItem(hDlg, IDC_OPTIONS_ASSOCIATE_PRINT);
    hWndMarginLeft=GetDlgItem(hDlg, IDC_OPTIONS_EDITMARGIN_LEFT);
    hWndMarginLeftSpin=GetDlgItem(hDlg, IDC_OPTIONS_EDITMARGIN_LEFT_SPIN);
    hWndMarginRight=GetDlgItem(hDlg, IDC_OPTIONS_EDITMARGIN_RIGHT);
    hWndMarginRightSpin=GetDlgItem(hDlg, IDC_OPTIONS_EDITMARGIN_RIGHT_SPIN);
    hWndDelimiters=GetDlgItem(hDlg, IDC_OPTIONS_DELIMITERS);

    SendMessage(hWndAssociateOpen, BM_SETCHECK, (BOOL)(dwFileTypesAssociated & AE_OPEN), 0);
    SendMessage(hWndAssociateEdit, BM_SETCHECK, (BOOL)(dwFileTypesAssociated & AE_EDIT), 0);
    SendMessage(hWndAssociatePrint, BM_SETCHECK, (BOOL)(dwFileTypesAssociated & AE_PRINT), 0);
    EnableWindow(hWndFileTypesOpen, (BOOL)(dwFileTypesAssociated & AE_OPEN));
    EnableWindow(hWndFileTypesEdit, (BOOL)(dwFileTypesAssociated & AE_EDIT));
    EnableWindow(hWndFileTypesPrint, (BOOL)(dwFileTypesAssociated & AE_PRINT));
    SendMessage(hWndFileTypesOpen, EM_LIMITTEXT, (WPARAM)MAX_PATH, 0);
    SendMessage(hWndFileTypesEdit, EM_LIMITTEXT, (WPARAM)MAX_PATH, 0);
    SendMessage(hWndFileTypesPrint, EM_LIMITTEXT, (WPARAM)MAX_PATH, 0);

    SendMessage(hWndDelimiters, EM_LIMITTEXT, (WPARAM)DELIMITERS_SIZE, 0);

    SendMessage(hWndMarginLeftSpin, UDM_SETBUDDY, (WPARAM)hWndMarginLeft, 0);
    SendMessage(hWndMarginLeftSpin, UDM_SETRANGE, 0, (LPARAM)MAKELONG(999, 0));
    SendMessage(hWndMarginRightSpin, UDM_SETBUDDY, (WPARAM)hWndMarginRight, 0);
    SendMessage(hWndMarginRightSpin, UDM_SETRANGE, 0, (LPARAM)MAKELONG(999, 0));

    SetWindowText(hWndFileTypesOpen, szFileTypesOpen);
    SetWindowText(hWndFileTypesEdit, szFileTypesEdit);
    SetWindowText(hWndFileTypesPrint, szFileTypesPrint);

    SetDlgItemInt(hDlg, IDC_OPTIONS_EDITMARGIN_LEFT, LOWORD(dwEditMargins), FALSE);
    SetDlgItemInt(hDlg, IDC_OPTIONS_EDITMARGIN_RIGHT, HIWORD(dwEditMargins), FALSE);

    EscapeDataToEscapeStringW(wszDelimiters, (wchar_t *)buf);
    WideCharToMultiByte(CP_ACP, 0, (wchar_t *)buf, -1, buf2, BUFFER_SIZE, NULL, NULL);
    SetWindowText(hWndDelimiters, buf2);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_OPTIONS_ASSOCIATE_OPEN)
    {
      bState=SendMessage(hWndAssociateOpen, BM_GETCHECK, 0, 0);
      EnableWindow(hWndFileTypesOpen, bState);
      if (!bState) SetWindowText(hWndFileTypesOpen, szFileTypesOpen);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OPTIONS_ASSOCIATE_EDIT)
    {
      bState=SendMessage(hWndAssociateEdit, BM_GETCHECK, 0, 0);
      EnableWindow(hWndFileTypesEdit, bState);
      if (!bState) SetWindowText(hWndFileTypesEdit, szFileTypesEdit);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OPTIONS_ASSOCIATE_PRINT)
    {
      bState=SendMessage(hWndAssociatePrint, BM_GETCHECK, 0, 0);
      EnableWindow(hWndFileTypesPrint, bState);
      if (!bState) SetWindowText(hWndFileTypesPrint, szFileTypesPrint);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_OPTIONS_RESET_DELIMITERS)
    {
      EscapeDataToEscapeStringW(WORD_DELIMITERSW, (wchar_t *)buf);
      WideCharToMultiByte(CP_ACP, 0, (wchar_t *)buf, -1, buf2, BUFFER_SIZE, NULL, NULL);
      SetDlgItemText(hDlg, IDC_OPTIONS_DELIMITERS, buf2);
      return TRUE;
    }
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (((NMHDR *)lParam)->code == PSN_APPLY)
    {
      char szModule[MAX_PATH];
      char szWindowText[MAX_PATH];
      BOOL bShellRefresh=FALSE;
      int a;
      int b;

      //Associations
      GetModuleFileName(NULL, szModule, MAX_PATH);

      bState=SendMessage(hWndAssociateOpen, BM_GETCHECK, 0, 0);
      if (bState && !(dwFileTypesAssociated & AE_OPEN))
      {
        GetWindowText(hWndFileTypesOpen, szFileTypesOpen, MAX_PATH);
        AssociateFileTypes(szFileTypesOpen, szModule, AE_OPEN|AE_ASSOCIATE);
        dwFileTypesAssociated|=AE_OPEN;
        bShellRefresh=TRUE;
      }
      else if (bState && (dwFileTypesAssociated & AE_OPEN))
      {
        GetWindowText(hWndFileTypesOpen, szWindowText, MAX_PATH);
        if (lstrcmpi(szWindowText, szFileTypesOpen))
        {
          AssociateFileTypes(szFileTypesOpen, szModule, AE_OPEN|AE_DEASSOCIATE);
          AssociateFileTypes(szWindowText, szModule, AE_OPEN|AE_ASSOCIATE);
          lstrcpy(szFileTypesOpen, szWindowText);
        }
        bShellRefresh=TRUE;
      }
      else if (!bState && (dwFileTypesAssociated & AE_OPEN))
      {
        AssociateFileTypes(szFileTypesOpen, szModule, AE_OPEN|AE_DEASSOCIATE);
        dwFileTypesAssociated&=~AE_OPEN;
        bShellRefresh=TRUE;
      }

      bState=SendMessage(hWndAssociateEdit, BM_GETCHECK, 0, 0);
      if (bState && !(dwFileTypesAssociated & AE_EDIT))
      {
        GetWindowText(hWndFileTypesEdit, szFileTypesEdit, MAX_PATH);
        AssociateFileTypes(szFileTypesEdit, szModule, AE_EDIT|AE_ASSOCIATE);
        dwFileTypesAssociated|=AE_EDIT;
        bShellRefresh=TRUE;
      }
      else if (bState && (dwFileTypesAssociated & AE_EDIT))
      {
        GetWindowText(hWndFileTypesEdit, szWindowText, MAX_PATH);
        if (lstrcmpi(szWindowText, szFileTypesEdit))
        {
          AssociateFileTypes(szFileTypesEdit, szModule, AE_EDIT|AE_DEASSOCIATE);
          AssociateFileTypes(szWindowText, szModule, AE_EDIT|AE_ASSOCIATE);
          lstrcpy(szFileTypesEdit, szWindowText);
        }
        bShellRefresh=TRUE;
      }
      else if (!bState && (dwFileTypesAssociated & AE_EDIT))
      {
        AssociateFileTypes(szFileTypesEdit, szModule, AE_EDIT|AE_DEASSOCIATE);
        dwFileTypesAssociated&=~AE_EDIT;
        bShellRefresh=TRUE;
      }

      bState=SendMessage(hWndAssociatePrint, BM_GETCHECK, 0, 0);
      if (bState && !(dwFileTypesAssociated & AE_PRINT))
      {
        GetWindowText(hWndFileTypesPrint, szFileTypesPrint, MAX_PATH);
        AssociateFileTypes(szFileTypesPrint, szModule, AE_PRINT|AE_ASSOCIATE);
        dwFileTypesAssociated|=AE_PRINT;
        bShellRefresh=TRUE;
      }
      else if (bState && (dwFileTypesAssociated & AE_PRINT))
      {
        GetWindowText(hWndFileTypesPrint, szWindowText, MAX_PATH);
        if (lstrcmpi(szWindowText, szFileTypesPrint))
        {
          AssociateFileTypes(szFileTypesPrint, szModule, AE_PRINT|AE_DEASSOCIATE);
          AssociateFileTypes(szWindowText, szModule, AE_PRINT|AE_ASSOCIATE);
          lstrcpy(szFileTypesPrint, szWindowText);
        }
        bShellRefresh=TRUE;
      }
      else if (!bState && (dwFileTypesAssociated & AE_PRINT))
      {
        AssociateFileTypes(szFileTypesPrint, szModule, AE_PRINT|AE_DEASSOCIATE);
        dwFileTypesAssociated&=~AE_PRINT;
        bShellRefresh=TRUE;
      }
      if (bShellRefresh) SHChangeNotify(SHCNE_ASSOCCHANGED, 0, 0, 0);

      //Margins
      a=GetDlgItemInt(hDlg, IDC_OPTIONS_EDITMARGIN_LEFT, NULL, FALSE);
      b=GetDlgItemInt(hDlg, IDC_OPTIONS_EDITMARGIN_RIGHT, NULL, FALSE);
      if (dwEditMargins != (DWORD)MAKELONG(a, b))
        dwEditMargins=MAKELONG(a, b);

      //Word delimiters
      a=GetDlgItemText(hDlg, IDC_OPTIONS_DELIMITERS, buf, BUFFER_SIZE);
      MultiByteToWideChar(CP_ACP, 0, buf, ++a, (wchar_t *)buf2, BUFFER_SIZE / sizeof(wchar_t));
      EscapeStringToEscapeDataW((wchar_t *)buf2, wszDelimiters);
    }
  }
  return FALSE;
}


//// About

BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    SetDlgItemText(hDlg, IDC_ABOUT_VERSION, APP_ABOUT_VERSION);
    SetDlgItemText(hDlg, IDC_ABOUT_HOMEPAGE, APP_HOMEPAGE);
    SetDlgItemText(hDlg, IDC_ABOUT_EMAIL_KUZNETSOV, APP_ABOUT_EMAIL_KUZNETSOV);
    SetDlgItemText(hDlg, IDC_ABOUT_ICQ_KUZNETSOV, APP_ABOUT_ICQ_KUZNETSOV);
    SetDlgItemText(hDlg, IDC_ABOUT_EMAIL_SHENGALTS, APP_ABOUT_EMAIL_SHENGALTS);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
    else if (LOWORD(wParam) == IDC_ABOUT_HOMEPAGE)
    {
      ShellExecute(GetParent(hDlg), "open", APP_HOMEPAGE, NULL, NULL, SW_MAXIMIZE);
      return TRUE;
    }
  }
  return FALSE;
}


//// Other functions

char* GetCommandLineParamsA()
{
  char *lpCmdLine=GetCommandLine();

  if (*lpCmdLine++ == '\"')
    while (*lpCmdLine != '\"' && *lpCmdLine != '\0') ++lpCmdLine;
  else
    while (*lpCmdLine != ' ' && *lpCmdLine != '\0') ++lpCmdLine;
  if (*lpCmdLine != '\0')
    while (*++lpCmdLine == ' ');

  return lpCmdLine;
}

int GetCommandLineArgA(char *lpCmdLine, char *szArg, char **lpNextArg)
{
  char *pCount=lpCmdLine;
  char *pArg=szArg;

  while (*pCount == ' ')
    ++pCount;

  while (*pCount != '\0' && *pCount != ' ')
  {
    for (; *pCount != '\"' && *pCount != ' ' && *pCount != '\0'; ++pArg, ++pCount)
      if (szArg) *pArg=*pCount;

    if (*pCount == '\"')
    {
      for (++pCount; *pCount != '\"' && *pCount != '\0'; ++pArg, ++pCount)
        if (szArg) *pArg=*pCount;
      if (*pCount == '\"') ++pCount;
    }
  }
  if (szArg) *pArg='\0';

  if (lpNextArg)
    for (*lpNextArg=pCount; **lpNextArg == ' '; ++*lpNextArg);

  return (pArg - szArg);
}

void GetAssociatedIconA(char *pExt, char *szFile, int *nIconIndex, HICON *phiconLarge, HICON *phiconSmall)
{
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;
  int nIndex=0;
  int i;
  int j;

  if (pExt=GetFileExtA(pExt))
  {
    if (RegOpenKeyEx(HKEY_CLASSES_ROOT, pExt - 1, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
    {
      dwSize=BUFFER_SIZE;
      RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)buf, &dwSize);
      RegCloseKey(hKey);
  
      wsprintf(buf2, "%s\\DefaultIcon", buf);
  
      if (RegOpenKeyEx(HKEY_CLASSES_ROOT, buf2, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
      {
        dwSize=BUFFER_SIZE;
        RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)buf2, &dwSize);
        RegCloseKey(hKey);
  
        for (i=0, j=0; buf2[i]; ++i)
          if (buf2[i] != '\"') buf[j++]=buf2[i];
        buf[j]='\0';

        for (i=lstrlen(buf) - 1; i > 0; --i)
        {
          if (buf[i] == ',')
          {
            buf[i]='\0';
            nIndex=xatoi(buf + i + 1);
            if (nIndex == -1) nIndex=0;
            break;
          }
          else if (buf[i] == '-' || (buf[i] >= '0' && buf[i] <= '9'))
            continue;
          break;
        }
        ExpandEnvironmentStrings(buf, buf2, BUFFER_SIZE);
        ExtractIconEx(buf2, nIndex, phiconLarge, phiconSmall, 1);

        if (szFile) lstrcpy(szFile, buf2);
        if (nIconIndex) *nIconIndex=nIndex;
        return;
      }
    }
  }
  if (szFile) szFile[0]='\0';
  if (nIconIndex) *nIconIndex=0;
  if (phiconLarge) *phiconLarge=0;
  if (phiconSmall) *phiconSmall=0;
}

void GetAssociatedIconW(wchar_t *pExt, wchar_t *wszFile, int *nIconIndex, HICON *phiconLarge, HICON *phiconSmall)
{
  wchar_t *wbuf=(wchar_t *)buf;
  wchar_t *wbuf2=(wchar_t *)buf2;
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;
  int nIndex;
  int i;
  int j;

  if (pExt=GetFileExtW(pExt))
  {
    if (RegOpenKeyExW(HKEY_CLASSES_ROOT, pExt - 1, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
    {
      dwSize=BUFFER_SIZE / sizeof(wchar_t);
      RegQueryValueExW(hKey, L"", NULL, &dwType, (LPBYTE)wbuf, &dwSize);
      RegCloseKey(hKey);
  
      wsprintfW(wbuf2, L"%s\\DefaultIcon", wbuf);
  
      if (RegOpenKeyExW(HKEY_CLASSES_ROOT, wbuf2, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
      {
        dwSize=BUFFER_SIZE / sizeof(wchar_t);
        RegQueryValueExW(hKey, L"", NULL, &dwType, (LPBYTE)wbuf2, &dwSize);
        RegCloseKey(hKey);
  
        for (i=0, j=0; wbuf2[i]; ++i)
          if (wbuf2[i] != '\"') wbuf[j++]=wbuf2[i];
        wbuf[j]='\0';

        for (i=lstrlenW(wbuf) - 1; i > 0; --i)
        {
          if (wbuf[i] == ',')
          {
            wbuf[i]='\0';
            WideCharToMultiByte(CP_ACP, 0, wbuf + i + 1, -1, (char *)wbuf2, BUFFER_SIZE, 0, 0);
            nIndex=xatoi((char *)wbuf2);
            if (nIndex == -1) nIndex=0;
            break;
          }
          else if (wbuf[i] == '-' || (wbuf[i] >= '0' && wbuf[i] <= '9'))
            continue;
          break;
        }
        ExpandEnvironmentStringsW(wbuf, wbuf2, BUFFER_SIZE / sizeof(wchar_t));
        (*ExtractIconExWPtr)(wbuf2, nIndex, phiconLarge, phiconSmall, 1);

        if (wszFile) lstrcpyW(wszFile, wbuf2);
        if (nIconIndex) *nIconIndex=nIndex;
        return;
      }
    }
  }
  if (wszFile) wszFile[0]='\0';
  if (nIconIndex) *nIconIndex=0;
  if (phiconLarge) *phiconLarge=0;
  if (phiconSmall) *phiconSmall=0;
}

void AssociateFileTypes(char *pFileTypes, char *pModule, DWORD Flags)
{
  char szAssocKey[MAX_PATH];
  char szTypeKey[16];
  char szExt[16];
  char *pExtStart=pFileTypes;
  char *pExtEnd=pFileTypes;
  HKEY hKey;
  DWORD dwType;
  DWORD dwSize;

  if (Flags & AE_OPEN) lstrcpy(szTypeKey, "Open");
  else if (Flags & AE_EDIT) lstrcpy(szTypeKey, "Edit");
  else if (Flags & AE_PRINT) lstrcpy(szTypeKey, "Print");

  while (1)
  {
    if (pExtEnd=strchr(pExtStart, ';'))
      lstrcpyn(buf, pExtStart, pExtEnd - pExtStart + 1);
    else
      lstrcpy(buf, pExtStart);
    if (!*buf) goto NextExt;
    wsprintf(szExt, ".%s", buf);

    if (RegCreateKeyEx(HKEY_CLASSES_ROOT, szExt, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) == ERROR_SUCCESS)
    {
      dwSize=MAX_PATH;
      if (RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)szAssocKey, &dwSize) != ERROR_SUCCESS || !*szAssocKey)
      {
        if (Flags & AE_ASSOCIATE)
        {
          wsprintf(szAssocKey, "%sfile", szExt + 1);
          RegSetValueEx(hKey, "", 0, REG_SZ, (LPBYTE)szAssocKey, lstrlen(szAssocKey) + 1);
        }
        else if (Flags & AE_DEASSOCIATE)
        {
          RegCloseKey(hKey);
          return;
        }
      }
      RegCloseKey(hKey);

      //Associate icon
      if (Flags & AE_OPEN)
      {
        wsprintf(buf2, "%s\\DefaultIcon", szAssocKey);
  
        if (RegCreateKeyEx(HKEY_CLASSES_ROOT, buf2, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) == ERROR_SUCCESS)
        {
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)buf2, &dwSize) != ERROR_SUCCESS || !*buf2)
          {
            if (Flags & AE_ASSOCIATE) RegSetValueEx(hKey, "", 0, REG_SZ, (LPBYTE)pModule, lstrlen(pModule) + 1);
          }
          else if (Flags & AE_DEASSOCIATE)
          {
            if (!lstrcmpi(buf2, pModule)) RegDeleteValue(hKey, "");
          }
          RegCloseKey(hKey);
        }
      }

      //Associate command
      wsprintf(buf2, "%s\\Shell\\%s\\Command", szAssocKey, szTypeKey);
      wsprintf(buf, "\"%s\"%s \"%%1\"", pModule, (Flags & AE_PRINT)?" /p":"");

      if (RegCreateKeyEx(HKEY_CLASSES_ROOT, buf2, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, NULL) == ERROR_SUCCESS)
      {
        if (Flags & AE_ASSOCIATE)
        {
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)buf2, &dwSize) == ERROR_SUCCESS)
          {
            if (*buf2 && lstrcmpi(buf, buf2))
            {
              if (RegQueryValueEx(hKey, "AkelUndo", NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
                RegSetValueEx(hKey, "AkelUndo", 0, dwType, (LPBYTE)buf2, dwSize);
            }
          }
          RegSetValueEx(hKey, "", 0, REG_SZ, (LPBYTE)buf, lstrlen(buf) + 1);
        }
        else if (Flags & AE_DEASSOCIATE)
        {
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "AkelUndo", NULL, &dwType, (LPBYTE)buf2, &dwSize) == ERROR_SUCCESS)
          {
            if (*buf2) RegSetValueEx(hKey, "", 0, dwType, (LPBYTE)buf2, dwSize);
            RegDeleteValue(hKey, "AkelUndo");
          }
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "", NULL, &dwType, (LPBYTE)buf2, &dwSize) == ERROR_SUCCESS)
          {
            if (!*buf2 || !lstrcmpi(buf, buf2))
            {
              RegCloseKey(hKey);
              wsprintf(buf2, "%s\\Shell\\%s", szAssocKey, szTypeKey);
              if (RegOpenKeyEx(HKEY_CLASSES_ROOT, buf2, 0, KEY_ENUMERATE_SUB_KEYS, &hKey) == ERROR_SUCCESS)
              {
                RegDeleteKey(hKey, "Command");
                RegCloseKey(hKey);
                wsprintf(buf2, "%s\\Shell", szAssocKey);
                if (RegOpenKeyEx(HKEY_CLASSES_ROOT, buf2, 0, KEY_ENUMERATE_SUB_KEYS, &hKey) == ERROR_SUCCESS)
                {
                  RegDeleteKey(hKey, szTypeKey);
                  dwSize=BUFFER_SIZE;
                  if (RegEnumKeyEx(hKey, 0, buf2, &dwSize, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
                  {
                    RegCloseKey(hKey);
                    if (RegOpenKeyEx(HKEY_CLASSES_ROOT, szAssocKey, 0, KEY_ENUMERATE_SUB_KEYS, &hKey) == ERROR_SUCCESS)
                      RegDeleteKey(hKey, "Shell");
                  }
                }
              }
            }
          }
        }
        RegCloseKey(hKey);
      }
    }

    //Check off "Always open with this program"
    if (Flags & AE_OPEN)
    {
      wsprintf(buf, "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\%s", szExt);

      if (RegOpenKeyEx(HKEY_CURRENT_USER, buf, 0, KEY_QUERY_VALUE|KEY_SET_VALUE, &hKey) == ERROR_SUCCESS)
      {
        if (Flags & AE_ASSOCIATE)
        {
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "Application", NULL, &dwType, (LPBYTE)buf, &dwSize) == ERROR_SUCCESS)
          {
            if (*buf) RegSetValueEx(hKey, "AkelUndo", 0, dwType, (LPBYTE)buf, dwSize);
            RegDeleteValue(hKey, "Application");
          }
        }
        else if (Flags & AE_DEASSOCIATE)
        {
          dwSize=BUFFER_SIZE;
          if (RegQueryValueEx(hKey, "AkelUndo", NULL, &dwType, (LPBYTE)buf, &dwSize) == ERROR_SUCCESS)
          {
            if (*buf) RegSetValueEx(hKey, "Application", 0, dwType, (LPBYTE)buf, dwSize);
            RegDeleteValue(hKey, "AkelUndo");
          }
        }
      }
    }
    NextExt:
    pExtStart=pExtEnd + 1;
    if (!pExtEnd) break;
    if (!*pExtStart) break;
  }
}

BOOL SaveChanged()
{
  int nChoice;

  if (bModified)
  {
    API_LoadString(hLangLib, MSG_DOCUMENT_CHANGED, buf, BUFFER_SIZE);
    nChoice=MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_YESNOCANCEL|MB_ICONEXCLAMATION);

    if (nChoice == IDYES)
    {
      if (!DoFileSave()) return FALSE;
    }
    else if (nChoice == IDCANCEL) return FALSE;
  }
  return TRUE;
}

int SetLayout()
{
  int nLayout=LOWORD(GetKeyboardLayout(0));
  if (nLayout != wSystemLanguage && !ActivateKeyboardLayout((HKL)wSystemLanguage, 0))
    nLayout=0;
  return nLayout;
}

void RestoreLayout(int nLayout)
{
  if (nLayout) ActivateKeyboardLayout((HKL)nLayout, 0);
}

void SetChosenFont(HWND hWnd, LOGFONT logfont)
{
  CHARFORMAT cfmt;

  cfmt.cbSize=sizeof(CHARFORMAT);
  cfmt.dwMask=CFM_SIZE|CFM_CHARSET|CFM_FACE|CFM_BOLD|CFM_ITALIC|CFM_UNDERLINE|CFM_STRIKEOUT;
  cfmt.yHeight=PixelsToTwips(hWnd, logfont.lfHeight);
  cfmt.bCharSet=logfont.lfCharSet;
  cfmt.bPitchAndFamily=logfont.lfPitchAndFamily;
  lstrcpy(cfmt.szFaceName, logfont.lfFaceName);
  cfmt.dwEffects=0;
  if (logfont.lfWeight >= FW_BOLD)
    cfmt.dwEffects|=CFE_BOLD;
  if (logfont.lfItalic)
    cfmt.dwEffects|=CFE_ITALIC;
  if (logfont.lfUnderline)
    cfmt.dwEffects|=CFE_UNDERLINE;
  if (logfont.lfStrikeOut)
    cfmt.dwEffects|=CFE_STRIKEOUT;

  SendMessage(hWnd, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cfmt);
}

void SetChosenFontColor(HWND hWnd, COLORREF crFont)
{
  CHARFORMAT cfmt;

  cfmt.cbSize=sizeof(CHARFORMAT);
  cfmt.dwMask=CFM_COLOR;
  cfmt.dwEffects=0;
  cfmt.crTextColor=crFont;

  SendMessage(hWnd, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cfmt);
}

void SetChosenFontSize(HWND hWnd, int nPixelHeight)
{
  CHARFORMAT cfmt;

  cfmt.cbSize=sizeof(CHARFORMAT);
  cfmt.dwMask=CFM_SIZE;
  cfmt.yHeight=PixelsToTwips(hWnd, nPixelHeight);

  SendMessage(hWnd, EM_SETCHARFORMAT, SCF_ALL, (LPARAM)&cfmt);
}

int PixelsToTwips(HWND hWnd, int nPixels)
{
  HDC hDC;
  int nTwips;

  hDC=GetDC(hWnd);
  if (nPixels < 0) nPixels=0 - nPixels;
  nTwips=nPixels*72*20/GetDeviceCaps(hDC, LOGPIXELSY);
  ReleaseDC(hWnd, hDC);

  return nTwips;
}

void SetModify(BOOL bState, BOOL bFirst)
{
  if (bFirst != TRUE && bModified == bState) return;
  bModified=bState;

  API_LoadString(hLangLib, STR_MODIFIED, buf, BUFFER_SIZE);
  SendMessage(hStatus, SB_SETTEXT, 1, (LPARAM)(bModified?buf:""));
}

void SetUnixNewLine(BOOL bState, BOOL bFirst)
{
  NMHDR nmhdr={hWndEdit, ID_EDIT, EN_SELCHANGE};

  if (bFirst != TRUE && bUnixNewLine == bState) return;
  bUnixNewLine=bState;

  SendMessage(hStatus, SB_SETTEXT, 3, (LPARAM)(bUnixNewLine?"Unix":"Win"));
  SendMessage(hMainWnd, WM_NOTIFY, ID_EDIT, (LPARAM)&nmhdr);
}

void SetInsertState(BOOL bState, BOOL bFirst)
{
  if (bFirst != TRUE && bInsertState == bState) return;
  bInsertState=bState;

  SendMessage(hStatus, SB_SETTEXT, 2, (LPARAM)(bInsertState?"Ovr":"Ins"));
}

void SetCodePage(int nCodePage, BOOL bFirst)
{
  CPINFOEX CPInfoEx;

  if (bFirst != TRUE && nCurrentCodePage == LOWORD(nCodePage) && bCurrentBOM == HIWORD(nCodePage)) return;
  nCurrentCodePage=LOWORD(nCodePage);
  bCurrentBOM=HIWORD(nCodePage);

  if (nCurrentCodePage == CP_UNICODE_UCS2_LE)
  {
    lstrcpy(buf, STR_UNICODE_UCS2_LE);
  }
  else if (nCurrentCodePage == CP_UNICODE_UCS2_BE)
  {
    lstrcpy(buf, STR_UNICODE_UCS2_BE);
  }
  else if (nCurrentCodePage == CP_UNICODE_UTF8)
  {
    lstrcpy(buf, STR_UNICODE_UTF8);
  }
  else
  {
    if (GetCPInfoExPtr && (*GetCPInfoExPtr)(nCurrentCodePage, 0, &CPInfoEx) && nCurrentCodePage == xatoi(CPInfoEx.CodePageName))
      lstrcpy(buf, CPInfoEx.CodePageName);
    else
      wsprintf(buf, "CP-%u", nCurrentCodePage);
  }
  SendMessage(hStatus, SB_SETTEXT, 4, (LPARAM)buf);
}

BOOL FileExists(char *fname)
{
  WIN32_FIND_DATA wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFile(fname, &wfd)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
}

int IsFile(char *fname)
{
  WIN32_FIND_DATA wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFile(fname, &wfd)) == INVALID_HANDLE_VALUE)
    return ERROR_INVALID_HANDLE;

  if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    return ERROR_FILE_NOT_FOUND;

  FindClose(hFind);
  return ERROR_SUCCESS;
}

BOOL IsPathFull(char *fpath)
{
  if (fpath[0] == '\\' && fpath[1] == '\\') return TRUE;
  if (fpath[1] != '\0' && fpath[1] == ':') return TRUE;
  return FALSE;
}

char* GetFileName(char *fname)
{
  int i;

  for (i=lstrlen(fname) - 1; i >= 0; --i)
    if (fname[i] == '\\') return (fname + i + 1);

  return fname;
}

char* GetBaseName(char *pFile, char *szBaseName)
{
  int nStart=lstrlen(pFile);
  int nEnd=nStart;

  while (nStart > 0 && pFile[nStart] != '\\')
  {
    if (pFile[nStart] == '.') nEnd=nStart;
    --nStart;
  }
  if (pFile[nStart] == '\\') ++nStart;
  lstrcpyn(szBaseName, pFile + nStart, nEnd - nStart + 1);

  return szBaseName;
}

char* GetFileExtA(char *fname)
{
  int i;

  for (i=lstrlen(fname) - 1; i >= 0; --i)
    if (fname[i] == '.') return (fname + i + 1);
    else if (fname[i] == '\\') break;

  return "";
}

wchar_t* GetFileExtW(wchar_t *fname)
{
  int i;

  for (i=lstrlenW(fname) - 1; i >= 0; --i)
    if (fname[i] == '.') return (fname + i + 1);
    else if (fname[i] == '\\') break;

  return L"";
}

void GetFileVersion(char *pFile, int *nMajor, int *nMinor, int *nRelease, int *nBuild)
{
  DWORD dwHandle;
  DWORD dwVerSize;
  void *pVerBuf;
  UINT uLen;
  VS_FIXEDFILEINFO ffi;
  VS_FIXEDFILEINFO *pffi=&ffi;
  *nMajor=*nMinor=*nRelease=*nBuild=0;

  if (dwVerSize=GetFileVersionInfoSize(pFile, &dwHandle))
  {
    if (pVerBuf=API_HeapAlloc(hHeap, 0, dwVerSize))
    {
      if (GetFileVersionInfo(pFile, dwHandle, dwVerSize, pVerBuf) && VerQueryValue(pVerBuf, "\\", (void **)&pffi, &uLen))
      {
        *nMajor=pffi->dwFileVersionMS / 0x00010000;
        *nMinor=pffi->dwFileVersionMS & 0x0000FFFF;
        *nRelease=pffi->dwFileVersionLS / 0x00010000;
        *nBuild=pffi->dwFileVersionLS & 0x0000FFFF;
      }
      API_HeapFree(hHeap, 0, pVerBuf);
    }
  }
}

int ParseExecCommand(char *pCommand, char *szBuffer, int nBufferSize)
{
  //%f == szCurrentFile, %e == szExeDir, %% == %
  int a;
  int b;
  char *pFile=szCurrentFile;
  char *pExeDir=szExeDir;
  int nFileLen=lstrlen(pFile);
  int nExeDirLen=lstrlen(pExeDir);

  for (a=0, b=0; pCommand[a] && (!szBuffer || b < nBufferSize); ++a, ++b)
  {
    if (pCommand[a] == '%')
    {
      if (pCommand[++a] == '%')
      {
        if (szBuffer) szBuffer[b]='%';
      }
      else if (pCommand[a] == 'f' || pCommand[a] == 'F')
      {
        if (!szBuffer || b + nFileLen <= nBufferSize)
        {
          if (szBuffer) lstrcpy(szBuffer + b, pFile);
          b+=nFileLen - 1;
        }
        else break;
      }
      else if (pCommand[a] == 'e' || pCommand[a] == 'E')
      {
        if (!szBuffer || b + nExeDirLen <= nBufferSize)
        {
          if (szBuffer) lstrcpy(szBuffer + b, pExeDir);
          b+=nExeDirLen - 1;
        }
        else break;
      }
      else break;
    }
    else if (szBuffer) szBuffer[b]=pCommand[a];
  }
  if (szBuffer && b < nBufferSize) szBuffer[b]='\0';
  return b;
}

void UpdateSize(int nWidth, int nHeight)
{
  int i;

  if (!bMDI)
  {
    MoveWindow(hWndEdit, 0, 0, nWidth, nHeight - (bStatusBar?nStatusHeight:0), TRUE);
  }
  else
  {
    i=nHeight - (bStatusBar?nStatusHeight:0) - (nTabView?TAB_HEIGHT:0);

    MoveWindow(hMdiClient, 0, (nTabView == TAB_VIEW_TOP)?TAB_HEIGHT:0, nWidth, i, TRUE);

    if (nTabView == TAB_VIEW_TOP)
      MoveWindow(hTab, 0, 0, nWidth, TAB_HEIGHT, TRUE);
    else if (nTabView == TAB_VIEW_BOTTOM)
      MoveWindow(hTab, 0, i, nWidth, TAB_HEIGHT, TRUE);
  }
}

void UpdateTitle(char *szFile)
{
  char *pFileName;

  //Get file name without path
  pFileName=GetFileName(szFile);

  if (!bMDI)
  {
    //Set caption of synchronizing window
    SetWindowText(hSyncWindow, szFile);

    //Set caption of main window
    wsprintf(buf, "%s - %s", pFileName, APP_SHORT_TITLE);
    SetWindowText(hMainWnd, buf);
  }
  else
  {
    TCITEM tcItem;
    HICON hIcon;
    int nCurSel;
    int i;

    //Set caption of MDI window
    SetWindowText(GetParent(hWndEdit), szFile);

    //Find file icon
    if (bOldWindows)
    {
      GetAssociatedIconA(pFileName, NULL, NULL, NULL, &hIcon);
    }
    else
    {
      MultiByteToWideChar(CP_ACP, 0, pFileName, -1, (wchar_t *)buf, BUFFER_SIZE);
      GetAssociatedIconW((wchar_t *)buf, NULL, NULL, NULL, &hIcon);
    }
    if (!hIcon) hIcon=hIconEmpty;

    //Set tab icon and text
    nCurSel=SendMessage(hTab, TCM_GETCURSEL, 0, 0);
    ImageList_Remove(hImageList, nCurSel);
    i=ImageList_AddIcon(hImageList, hIcon);

    tcItem.mask=TCIF_TEXT|TCIF_IMAGE;
    tcItem.pszText=pFileName;
    tcItem.iImage=i;
    SendMessage(hTab, TCM_SETITEM, nCurSel, (LPARAM)&tcItem);

    //Set MDI window icon
    if (lpWindowData->hIcon != hIconEmpty) DestroyIcon(lpWindowData->hIcon);
    lpWindowData->hIcon=hIcon;
    SendMessage(GetParent(hWndEdit), WM_SETICON, ICON_SMALL, (LPARAM)hIcon);
    DrawMenuBar(hMainWnd);
  }
}

void FreeMemorySearch()
{
  if (wszFind_orig)
  {
    API_HeapFree(hHeap, 0, (LPVOID)wszFind_orig);
    if (wszFind_orig == wszFind)
      wszFind=NULL;
    wszFind_orig=NULL;
  }
  if (wszReplace_orig)
  {
    API_HeapFree(hHeap, 0, (LPVOID)wszReplace_orig);
    if (wszReplace_orig == wszReplace)
      wszReplace=NULL;
    wszReplace_orig=NULL;
  }

  if (wszFind)
  {
    API_HeapFree(hHeap, 0, (LPVOID)wszFind);
    wszFind=NULL;
  }
  if (wszReplace)
  {
    API_HeapFree(hHeap, 0, (LPVOID)wszReplace);
    wszReplace=NULL;
  }
}

void FreeMemoryRecentFiles()
{
  if (szRecentNames)
  {
    API_HeapFree(hHeap, 0, (LPVOID)szRecentNames);
    szRecentNames=NULL;
  }
  if (szRecentPositions)
  {
    API_HeapFree(hHeap, 0, (LPVOID)szRecentPositions);
    szRecentPositions=NULL;
  }
  if (szRecentCodepages)
  {
    API_HeapFree(hHeap, 0, (LPVOID)szRecentCodepages);
    szRecentCodepages=NULL;
  }
}


//// API functions replacement

HMENU API_LoadMenu(HINSTANCE hLoadInstance, LPCTSTR lpMenuName)
{
  HMENU hResult;

  if (!(hResult=LoadMenu(hLoadInstance, lpMenuName)))
    if (hLoadInstance != hInstance)
      hResult=LoadMenu(hInstance, lpMenuName);

  return hResult;
}

HACCEL API_LoadAccelerators(HINSTANCE hLoadInstance, LPCTSTR lpTableName)
{
  HACCEL hResult;

  if (!(hResult=LoadAccelerators(hLoadInstance, lpTableName)))
    if (hLoadInstance != hInstance)
      hResult=LoadAccelerators(hInstance, lpTableName);

  return hResult;
}

HICON API_LoadIcon(HINSTANCE hLoadInstance, LPCTSTR lpIconName)
{
  HICON hResult;

  if (!(hResult=LoadIcon(hLoadInstance, lpIconName)))
    if (hLoadInstance != hInstance)
      hResult=LoadIcon(hInstance, lpIconName);

  return hResult;
}

HANDLE API_LoadImage(HINSTANCE hLoadInstance, LPCTSTR lpszName, UINT uType, int cxDesired, int cyDesired, UINT fuLoad)
{
  HANDLE hResult;

  if (!(hResult=LoadImage(hLoadInstance, lpszName, uType, cxDesired, cyDesired, fuLoad)))
    if (hLoadInstance != hInstance)
      hResult=LoadImage(hInstance, lpszName, uType, cxDesired, cyDesired, fuLoad);

  return hResult;
}

int API_LoadString(HINSTANCE hLoadInstance, UINT uID, LPTSTR lpBuffer, int nBufferMax)
{
  int nResult;

  if (!(nResult=LoadString(hLoadInstance, uID, lpBuffer, nBufferMax)))
    if (hLoadInstance != hInstance)
      if (!(nResult=LoadString(hInstance, uID, lpBuffer, nBufferMax)))
        lpBuffer[0]='\0';

  return nResult;
}

HWND API_CreateDialogA(HINSTANCE hLoadInstance, LPCTSTR lpTemplateName, HWND hWndParent, DLGPROC lpDialogFunc)
{
  HWND hResult;

  if (!(hResult=CreateDialogA(hLoadInstance, lpTemplateName, hWndParent, lpDialogFunc)))
    if (hLoadInstance != hInstance)
      hResult=CreateDialogA(hInstance, lpTemplateName, hWndParent, lpDialogFunc);

  return hResult;
}

HWND API_CreateDialogW(HINSTANCE hLoadInstance, LPCWSTR lpTemplateName, HWND hWndParent, DLGPROC lpDialogFunc)
{
  HWND hResult;

  if (!(hResult=CreateDialogW(hLoadInstance, lpTemplateName, hWndParent, lpDialogFunc)))
    if (hLoadInstance != hInstance)
      hResult=CreateDialogW(hInstance, lpTemplateName, hWndParent, lpDialogFunc);

  return hResult;
}

INT_PTR API_DialogBoxA(HINSTANCE hLoadInstance, LPCTSTR lpTemplateName, HWND hWndParent, DLGPROC lpDialogFunc)
{
  INT_PTR nResult;

  if ((nResult=DialogBoxA(hLoadInstance, lpTemplateName, hWndParent, lpDialogFunc)) == -1)
    if (hLoadInstance != hInstance)
      nResult=DialogBoxA(hInstance, lpTemplateName, hWndParent, lpDialogFunc);

  return nResult;
}

INT_PTR API_DialogBoxW(HINSTANCE hLoadInstance, LPCWSTR lpTemplateName, HWND hWndParent, DLGPROC lpDialogFunc)
{
  INT_PTR nResult;

  if ((nResult=DialogBoxW(hLoadInstance, lpTemplateName, hWndParent, lpDialogFunc)) == -1)
    if (hLoadInstance != hInstance)
      nResult=DialogBoxW(hInstance, lpTemplateName, hWndParent, lpDialogFunc);

  return nResult;
}

LPVOID API_HeapAlloc(HANDLE hHeap, DWORD dwFlags, SIZE_T dwBytes)
{
  LPVOID lpResult;

  if (!(lpResult=HeapAlloc(hHeap, dwFlags, dwBytes)))
  {
    API_LoadString(hLangLib, MSG_ERROR_NOT_ENOUGH_MEMORY, buf, BUFFER_SIZE);
    MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
  }

  return lpResult;
}

BOOL API_HeapFree(HANDLE hHeap, DWORD dwFlags, LPVOID lpMem)
{
  BOOL bResult;

  bResult=HeapFree(hHeap, dwFlags, lpMem);

  return bResult;
}

HANDLE API_CreateFile(LPCTSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile)
{
  HANDLE hResult;

  if ((hResult=CreateFile(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile)) == INVALID_HANDLE_VALUE)
  {
    API_LoadString(hLangLib, MSG_ERROR_CANNOT_OPEN_FILE, buf, BUFFER_SIZE);
    wsprintf(buf2, buf, lpFileName);
    MessageBox(hMainWnd, buf2, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
  }

  return hResult;
}

BOOL API_ReadFile(HANDLE hFile, LPVOID lpBuffer, DWORD nNumberOfBytesToRead, LPDWORD lpNumberOfBytesRead, LPOVERLAPPED lpOverlapped)
{
  BOOL bResult;

  if (!(bResult=ReadFile(hFile, lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, lpOverlapped)))
  {
    API_LoadString(hLangLib, MSG_ERROR_IO, buf, BUFFER_SIZE);
    MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
  }

  return bResult;
}

BOOL API_WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)
{
  BOOL bResult;

  if (!(bResult=WriteFile(hFile, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped)))
  {
    API_LoadString(hLangLib, MSG_ERROR_IO, buf, BUFFER_SIZE);
    MessageBox(hMainWnd, buf, APP_SHORT_TITLE, MB_OK|MB_ICONERROR);
  }

  return bResult;
}

BOOL API_SetWindowTextW(HWND hWnd, LPCWSTR lpString)
{
  BOOL bResult=FALSE;
  char *pData;
  int nLenW;

  if (bOldWindows)
  {
    nLenW=wcslen(lpString) + 1;

    if (pData=(char *)API_HeapAlloc(hHeap, 0, nLenW + 1))
    {
      WideCharToMultiByte(CP_ACP, 0, lpString, nLenW, pData, nLenW, NULL, NULL);
      bResult=SetWindowTextA(hWnd, pData);
      API_HeapFree(hHeap, 0, (LPVOID)pData);
    }
  }
  else
    bResult=SetWindowTextW(hWnd, lpString);

  return bResult;
}

int API_lstrlenW(LPCWSTR lpString)
{
  int nResult;

  if (bOldWindows)
    nResult=wcslen(lpString);
  else
    nResult=lstrlenW(lpString);

  return nResult;
}

int API_lstrcmpW(LPCWSTR lpString1, LPCWSTR lpString2)
{
  int nResult;

  if (bOldWindows)
    nResult=wcscmp(lpString1, lpString2);
  else
    nResult=lstrcmpW(lpString1, lpString2);

  return nResult;
}

int API_lstrcmpiW(LPCWSTR lpString1, LPCWSTR lpString2)
{
  int nResult=0;
  char *pData1;
  char *pData2;
  int nLen1;
  int nLen2;

  if (bOldWindows)
  {
    nLen1=wcslen(lpString1) + 1;
    nLen2=wcslen(lpString2) + 1;

    if (pData1=(char *)API_HeapAlloc(hHeap, 0, nLen1 + 1))
    {
      if (pData2=(char *)API_HeapAlloc(hHeap, 0, nLen2 + 1))
      {
        WideCharToMultiByte(CP_ACP, 0, lpString1, nLen1, pData1, nLen1, NULL, NULL);
        WideCharToMultiByte(CP_ACP, 0, lpString2, nLen2, pData2, nLen2, NULL, NULL);
        nResult=lstrcmpi(pData1, pData2);
        API_HeapFree(hHeap, 0, (LPVOID)pData2);
      }
      API_HeapFree(hHeap, 0, (LPVOID)pData1);
    }
  }
  else
    nResult=lstrcmpiW(lpString1, lpString2);

  return nResult;
}
